# -*- coding: utf-8 -*-


import numpy as np

from africanus.util.docs import DocstringTemplate
from africanus.util.numba import is_numba_type_none, generated_jit, jit
from africanus.util.requirements import requires_optional

try:
    from astropy.coordinates import CartesianRepresentation
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


@jit(nopython=True, nogil=True, cache=True)
def _create_phase_centre(phase_centre, dtype):
    return np.zeros((2,), dtype=dtype)


@jit(nopython=True, nogil=True, cache=True)
def _return_phase_centre(phase_centre, dtype):
    return phase_centre


@generated_jit(nopython=True, nogil=True, cache=True)
def radec_to_lmn(radec, phase_centre=None):
    dtype = radec.dtype

    if is_numba_type_none(phase_centre):
        _maybe_create_phase_centre = _create_phase_centre
    else:
        _maybe_create_phase_centre = _return_phase_centre

    def _radec_to_lmn_impl(radec, phase_centre=None):
        sources, components = radec.shape

        if radec.ndim != 2 or components != 2:
            raise ValueError("radec must have shape (source, 2)")

        lmn = np.empty(shape=(sources, 3), dtype=dtype)

        pc_ra, pc_dec = _maybe_create_phase_centre(phase_centre, dtype)
        sin_pc_dec = np.sin(pc_dec)
        cos_pc_dec = np.cos(pc_dec)

        for s in range(sources):
            ra_delta = radec[s, 0] - pc_ra
            sin_ra_delta = np.sin(ra_delta)
            cos_ra_delta = np.cos(ra_delta)

            sin_dec = np.sin(radec[s, 1])
            cos_dec = np.cos(radec[s, 1])

            lmn[s, 0] = l = cos_dec*sin_ra_delta  # noqa
            lmn[s, 1] = m = (sin_dec*cos_pc_dec -
                             cos_dec*sin_pc_dec*cos_ra_delta)
            lmn[s, 2] = np.sqrt(1.0 - l**2 - m**2)

        return lmn

    return _radec_to_lmn_impl


@generated_jit(nopython=True, nogil=True, cache=True)
def radec_to_lm(radec, phase_centre=None):
    dtype = radec.dtype

    if is_numba_type_none(phase_centre):
        _maybe_create_phase_centre = _create_phase_centre
    else:
        _maybe_create_phase_centre = _return_phase_centre

    def _radec_to_lm_impl(radec, phase_centre=None):
        sources, components = radec.shape

        if radec.ndim != 2 or components != 2:
            raise ValueError("radec must have shape (source, 2)")

        lm = np.empty(shape=(sources, 2), dtype=dtype)

        pc_ra, pc_dec = _maybe_create_phase_centre(phase_centre, dtype)
        sin_pc_dec = np.sin(pc_dec)
        cos_pc_dec = np.cos(pc_dec)

        for s in range(sources):
            da = radec[s, 0] - pc_ra
            sin_ra_delta = np.sin(da)
            cos_ra_delta = np.cos(da)

            sin_dec = np.sin(radec[s, 1])
            cos_dec = np.cos(radec[s, 1])

            lm[s, 0] = cos_dec*sin_ra_delta
            lm[s, 1] = sin_dec*cos_pc_dec - cos_dec*sin_pc_dec*cos_ra_delta

        return lm

    return _radec_to_lm_impl


@generated_jit(nopython=True, nogil=True, cache=True)
def lmn_to_radec(lmn, phase_centre=None):
    dtype = lmn.dtype

    if is_numba_type_none(phase_centre):
        _maybe_create_phase_centre = _create_phase_centre
    else:
        _maybe_create_phase_centre = _return_phase_centre

    def _lmn_to_radec_impl(lmn, phase_centre=None):
        if lmn.ndim != 2 or lmn.shape[1] != 3:
            raise ValueError("lmn must have shape (source, 3)")

        radec = np.empty(shape=(lmn.shape[0], 2), dtype=dtype)

        pc_ra, pc_dec = _maybe_create_phase_centre(phase_centre, dtype)
        sin_pc_dec = np.sin(pc_dec)
        cos_pc_dec = np.cos(pc_dec)

        for s in range(radec.shape[0]):
            l, m, n = lmn[s]

            radec[s, 1] = np.arcsin(m*cos_pc_dec + n*sin_pc_dec)
            radec[s, 0] = pc_ra + np.arctan(l / (n*cos_pc_dec - m*sin_pc_dec))

        return radec

    return _lmn_to_radec_impl


@generated_jit(nopython=True, nogil=True, cache=True)
def lm_to_radec(lm, phase_centre=None):
    dtype = lm.dtype

    if is_numba_type_none(phase_centre):
        _maybe_create_phase_centre = _create_phase_centre
    else:
        _maybe_create_phase_centre = _return_phase_centre

    def _lm_to_radec_impl(lm, phase_centre=None):
        if lm.ndim != 2 or lm.shape[1] != 2:
            raise ValueError("lm must have shape (source, 2)")

        radec = np.empty(shape=(lm.shape[0], 2), dtype=dtype)

        pc_ra, pc_dec = _maybe_create_phase_centre(phase_centre, dtype)
        sin_pc_dec = np.sin(pc_dec)
        cos_pc_dec = np.cos(pc_dec)

        for s in range(radec.shape[0]):
            l, m = lm[s]
            n = np.sqrt(1.0 - l**2 - m**2)

            radec[s, 1] = np.arcsin(m*cos_pc_dec + n*sin_pc_dec)
            radec[s, 0] = pc_ra + np.arctan(l / (n*cos_pc_dec - m*sin_pc_dec))

        return radec

    return _lm_to_radec_impl


@requires_optional("astropy", opt_import_error)
def astropy_radec_to_lmn(radec, phase_centre):
    """
    Astropy radec_to_lmn conversion, useful for testing.

    Parameters
    ----------
    radec : :class:`astropy.coordinates.SkyCoord`
        Sky coordinates
    phase_centre : :class:`astropy.coordinates.SkyCoord`
        Phase Centre

    Returns
    -------
    lmn : :class:`numpy.ndarray`
        lmn coordinates of shape :code:`(source, 3)`

    """
    # Transform radec relative to phase centre
    relative = radec.transform_to(phase_centre.skyoffset_frame())
    ret = relative.represent_as(CartesianRepresentation)

    # Rearrange astropy's coordinates into lmn convention
    result = np.empty((ret.x.value.shape[0], 3), dtype=ret.x.value.dtype)
    result[:, 0] = ret.y.value
    result[:, 1] = ret.z.value
    result[:, 2] = ret.x.value
    return result


RADEC_TO_LMN_DOCS = DocstringTemplate(r"""
Converts Right-Ascension/Declination coordinates in radians
to a Direction Cosine lm coordinates, relative to the Phase Centre.

.. math::
    :nowrap:

    \begin{eqnarray}
        & l =& \, \cos \, \delta  \sin \, \Delta \alpha  \\
        & m =& \, \sin \, \delta \cos \, \delta 0 -
                     \cos \delta \sin \delta 0 \cos \Delta \alpha \\
        & n =& \, \sqrt{1 - l^2 - m^2} - 1
    \end{eqnarray}

where :math:`\Delta \alpha = \alpha - \alpha 0` is the difference between
the Right Ascension of each coordinate and the phase centre and
:math:`\delta 0` is the Declination of the phase centre.

Parameters
----------
radec : $(array_type)
    radec coordinates of shape :code:`(coord, 2)`
    where Right-Ascension and Declination are in the
    last 2 components, respectively.
phase_centre : $(array_type), optional
    radec coordinates of the Phase Centre.
    Shape :code:`(2,)`

Returns
-------
$(array_type)
    lm Direction Cosines of shape :code:`(coord, $(lm_components))`
""")


LMN_TO_RADEC_DOCS = DocstringTemplate(r"""
Convert Direction Cosine lm coordinates to Right Ascension/Declination
coordinates in radians, relative to the Phase Centre.

.. math::
    :nowrap:

    \begin{eqnarray}
    & \delta = & \, \arcsin \left( m \cos \delta 0 +
                                 n \sin \delta 0 \right) \\
    & \alpha = & \, \arctan \left( \frac{l}{n \cos \delta 0 -
                                 m \sin \delta 0} \right) \\
    \end{eqnarray}

where :math:`\alpha` is the Right Ascension of each coordinate
and the phase centre and :math:`\delta 0`
is the Declination of the phase centre.

Parameters
----------
$(lm_name) : $(array_type)
    lm Direction Cosines of shape :code:`(coord, $(lm_components))`
phase_centre : $(array_type), optional
    radec coordinates of the Phase Centre.
    Shape :code:`(2,)`

Returns
-------
$(array_type)
    radec coordinates of shape :code:`(coord, 2)`
    where Right-Ascension and Declination are in the
    last 2 components, respectively.

""")

try:
    radec_to_lmn.__doc__ = RADEC_TO_LMN_DOCS.substitute(
                                lm_components="3",
                                array_type=":class:`numpy.ndarray`")
    radec_to_lm.__doc__ = RADEC_TO_LMN_DOCS.substitute(
                                lm_components="2",
                                array_type=":class:`numpy.ndarray`")
    lmn_to_radec.__doc__ = LMN_TO_RADEC_DOCS.substitute(
                                lm_name="lmn", lm_components="3",
                                array_type=":class:`numpy.ndarray`")
    lm_to_radec.__doc__ = LMN_TO_RADEC_DOCS.substitute(
                                lm_name="lm", lm_components="2",
                                array_type=":class:`numpy.ndarray`")

except AttributeError:
    pass



# -*- coding: utf-8 -*-

"""
Configure jax to use default 64 bit precision
https://github.com/google/jax/issues/216
"""


try:
    import jax.config
except ImportError:
    pass
else:
    jax.config.update("jax_enable_x64", True)



# -*- coding: utf-8 -*-


from africanus.dft.kernels import im_to_vis_docs, vis_to_im_docs
from africanus.dft.kernels import im_to_vis as np_im_to_vis
from africanus.dft.kernels import vis_to_im as np_vis_to_im

from africanus.util.docs import doc_tuple_to_str
from africanus.util.requirements import requires_optional

import numpy as np

try:
    import dask.array as da
except ImportError as e:
    dask_import_error = e
else:
    dask_import_error = None


def _im_to_vis_wrapper(image, uvw, lm, frequency, dtype_):
    return np_im_to_vis(image[0], uvw[0], lm[0][0],
                        frequency, dtype=dtype_)


@requires_optional('dask.array', dask_import_error)
def im_to_vis(image, uvw, lm, frequency, dtype=np.complex128):
    """ Dask wrapper for im_to_vis function """
    if lm.chunks[0][0] != lm.shape[0]:
        raise ValueError("lm chunks must match lm shape "
                         "on first axis")
    if image.chunks[0][0] != image.shape[0]:
        raise ValueError("Image chunks must match image "
                         "shape on first axis")
    if image.chunks[0][0] != lm.chunks[0][0]:
        raise ValueError("Image chunks and lm chunks must "
                         "match on first axis")
    if image.chunks[1] != frequency.chunks[0]:
        raise ValueError("Image chunks must match frequency "
                         "chunks on second axis")
    return da.core.blockwise(_im_to_vis_wrapper, ("row", "chan", "corr"),
                             image, ("source", "chan", "corr"),
                             uvw, ("row", "(u,v,w)"),
                             lm, ("source", "(l,m)"),
                             frequency, ("chan",),
                             dtype=dtype,
                             dtype_=dtype)


def _vis_to_im_wrapper(vis, uvw, lm, frequency, flags, dtype_):
    return np_vis_to_im(vis, uvw[0], lm[0],
                        frequency, flags,
                        dtype=dtype_)[None, :]


@requires_optional('dask.array', dask_import_error)
def vis_to_im(vis, uvw, lm, frequency, flags, dtype=np.float64):
    """ Dask wrapper for vis_to_im function """

    if vis.chunks[0] != uvw.chunks[0]:
        raise ValueError("Vis chunks and uvw chunks must "
                         "match on first axis")
    if vis.chunks[1] != frequency.chunks[0]:
        raise ValueError("Vis chunks must match frequency "
                         "chunks on second axis")
    if vis.chunks != flags.chunks:
        raise ValueError("Vis chunks must match flags "
                         "chunks on all axes")

    ims = da.core.blockwise(_vis_to_im_wrapper,
                            ("row", "source", "chan", "corr"),
                            vis, ("row", "chan", "corr"),
                            uvw, ("row", "(u,v,w)"),
                            lm, ("source", "(l,m)"),
                            frequency, ("chan",),
                            flags, ("row", "chan", "corr"),
                            adjust_chunks={"row": 1},
                            dtype=dtype,
                            dtype_=dtype)

    return ims.sum(axis=0)


im_to_vis.__doc__ = doc_tuple_to_str(im_to_vis_docs,
                                     [(":class:`numpy.ndarray`",
                                         ":class:`dask.array.Array`")])

vis_to_im.__doc__ = doc_tuple_to_str(vis_to_im_docs,
                                     [(":class:`numpy.ndarray`",
                                         ":class:`dask.array.Array`")])



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from africanus.model.shape import gaussian as np_gaussian


def test_gauss_shape():
    row = 10
    chan = 16

    shape_params = np.array([[.4, .3, .2],
                             [.4, .3, .2]])
    uvw = np.random.random((row, 3))
    freq = np.linspace(.856e9, 2*.856e9, chan)

    gauss_shape = np_gaussian(uvw, freq, shape_params)

    assert gauss_shape.shape == (shape_params.shape[0], row, chan)


def test_dask_gauss_shape():
    da = pytest.importorskip('dask.array')
    from africanus.model.shape.dask import gaussian as da_gaussian

    row_chunks = (5, 5)
    chan_chunks = (4, 4)

    row = sum(row_chunks)
    chan = sum(chan_chunks)

    shape_params = da.asarray([[.4, .3, .2],
                               [.4, .3, .2]])
    uvw = da.random.random((row, 3), chunks=(row_chunks, 3))
    freq = da.linspace(.856e9, 2*.856e9, chan, chunks=chan_chunks)
    da_gauss_shape = da_gaussian(uvw, freq, shape_params).compute()
    np_gauss_shape = np_gaussian(uvw.compute(),
                                 freq.compute(),
                                 shape_params.compute())

    assert_array_almost_equal(da_gauss_shape, np_gauss_shape)



# -*- coding: utf-8 -*-


from collections import namedtuple

from numba import types
import numpy as np

from africanus.averaging.time_and_channel_mapping import (row_mapper,
                                                          channel_mapper)
from africanus.util.docs import DocstringTemplate
from africanus.util.numba import is_numba_type_none, generated_jit, njit


def output_factory(present):
    """ Returns function creating an output if present """

    if present:
        def impl(rows, array):
            return np.zeros((rows,) + array.shape[1:], array.dtype)
    else:
        def impl(rows, array):
            return None

    return njit(nogil=True, cache=True)(impl)


def add_factory(present):
    """ Returns function for adding data to a bin """
    if present:
        def impl(output, orow, input, irow):
            output[orow] += input[irow]
    else:
        def impl(input, irow, output, orow):
            pass

    return njit(nogil=True, cache=True)(impl)


def comp_add_factory(present):
    """
    Returns function for adding data with components to a bin.
    Rows are assumed to be in the first dimension and
    components are assumed to be in the second
    """
    if present:
        def impl(output, orow, input, irow):
            for c in range(output.shape[1]):
                output[orow, c] += input[irow, c]
    else:
        def impl(input, irow, output, orow):
            pass

    return njit(nogil=True, cache=True)(impl)


def sigma_add_factory(have_sigma, have_weight):
    """
    Returns function for adding sigma values to a bin.
    Uses provided weights, else natural weights
    """
    if not have_sigma:
        def impl(out_sigma, out_weight_sum, orow, in_sigma, in_weight, irow):
            pass
    elif have_weight:
        def impl(out_sigma, out_weight_sum, orow, in_sigma, in_weight, irow):
            for c in range(out_sigma.shape[1]):
                out_sigma[orow, c] += (in_sigma[irow, c]**2 *
                                       in_weight[irow, c]**2)
                out_weight_sum[orow, c] += in_weight[irow, c]
    else:
        def impl(out_sigma, out_weight_sum, orow, in_sigma, in_weight, irow):
            for c in range(out_sigma.shape[1]):
                out_sigma[orow, c] += in_sigma[irow, c]**2
                out_weight_sum[orow, c] += in_weight[irow, c]

    return njit(nogil=True, cache=True)(impl)


def normaliser_factory(present):
    """ Returns function for normalising data in a bin """
    if present:
        def impl(data, row, bin_size):
            data[row] /= bin_size
    else:
        def impl(data, row, bin_size):
            pass

    return njit(nogil=True, cache=True)(impl)


def sigma_normaliser_factory(present):
    """ Returns function for normalising sigma in a bin """
    if present:
        def impl(sigma, row, weight_sum):
            for c in range(sigma.shape[1]):
                wt = weight_sum[row, c]

                if wt == 0.0:
                    continue

                sigma[row, c] = np.sqrt(sigma[row, c] / (wt**2))
    else:
        def impl(sigma, row, weight_sum):
            pass

    return njit(nogil=True, cache=True)(impl)


def matching_flag_factory(present):
    if present:
        def impl(flag_row, ri, out_flag_row, ro):
            return flag_row[ri] == out_flag_row[ro]
    else:
        def impl(flag_row, ri, out_flag_row, ro):
            return True

    return njit(nogil=True, cache=True)(impl)


_row_output_fields = ["antenna1", "antenna2", "time_centroid", "exposure",
                      "uvw", "weight", "sigma"]
RowAverageOutput = namedtuple("RowAverageOutput", _row_output_fields)


@generated_jit(nopython=True, nogil=True, cache=True)
def row_average(meta, ant1, ant2, flag_row=None,
                time_centroid=None, exposure=None, uvw=None,
                weight=None, sigma=None):

    have_flag_row = not is_numba_type_none(flag_row)
    have_uvw = not is_numba_type_none(uvw)
    have_time_centroid = not is_numba_type_none(time_centroid)
    have_exposure = not is_numba_type_none(exposure)
    have_weight = not is_numba_type_none(weight)
    have_sigma = not is_numba_type_none(sigma)

    flags_match = matching_flag_factory(have_flag_row)

    uvw_factory = output_factory(have_uvw)
    time_centroid_factory = output_factory(have_time_centroid)
    exposure_factory = output_factory(have_exposure)
    weight_factory = output_factory(have_weight)
    sigma_factory = output_factory(have_sigma)

    time_centroid_adder = add_factory(have_time_centroid)
    exposure_adder = add_factory(have_exposure)
    uvw_adder = comp_add_factory(have_uvw)
    weight_adder = comp_add_factory(have_weight)
    sigma_adder = sigma_add_factory(have_sigma, have_weight)

    uvw_normaliser = normaliser_factory(have_uvw)
    sigma_normaliser = sigma_normaliser_factory(have_sigma)
    time_centroid_normaliser = normaliser_factory(have_time_centroid)

    def impl(meta, ant1, ant2, flag_row=None,
             time_centroid=None, exposure=None, uvw=None,
             weight=None, sigma=None):

        out_rows = meta.time.shape[0]

        counts = np.zeros(out_rows, dtype=np.uint32)

        # These outputs are always present
        ant1_avg = np.empty(out_rows, ant1.dtype)
        ant2_avg = np.empty(out_rows, ant2.dtype)

        # Possibly present outputs for possibly present inputs
        uvw_avg = uvw_factory(out_rows, uvw)
        time_centroid_avg = time_centroid_factory(out_rows, time_centroid)
        exposure_avg = exposure_factory(out_rows, exposure)
        weight_avg = weight_factory(out_rows, weight)
        sigma_avg = sigma_factory(out_rows, sigma)
        sigma_weight_sum = sigma_factory(out_rows, sigma)

        # Iterate over input rows, accumulating into output rows
        for in_row, out_row in enumerate(meta.map):
            # Input and output flags must match in order for the
            # current row to contribute to these columns
            if flags_match(flag_row, in_row, meta.flag_row, out_row):
                uvw_adder(uvw_avg, out_row, uvw, in_row)
                weight_adder(weight_avg, out_row, weight, in_row)
                sigma_adder(sigma_avg, sigma_weight_sum, out_row,
                            sigma, weight, in_row)
                time_centroid_adder(time_centroid_avg, out_row,
                                    time_centroid, in_row)
                exposure_adder(exposure_avg, out_row, exposure, in_row)

                counts[out_row] += 1

            # Here we can simply assign because input_row baselines
            # should always match output row baselines
            ant1_avg[out_row] = ant1[in_row]
            ant2_avg[out_row] = ant2[in_row]

        # Normalise
        for out_row in range(out_rows):
            count = counts[out_row]

            if count > 0:
                uvw_normaliser(uvw_avg, out_row, count)
                time_centroid_normaliser(time_centroid_avg, out_row, count)
                sigma_normaliser(sigma_avg, out_row, sigma_weight_sum)

        return RowAverageOutput(ant1_avg, ant2_avg,
                                time_centroid_avg, exposure_avg, uvw_avg,
                                weight_avg, sigma_avg)

    return impl


def weight_sum_output_factory(present):
    """ Returns function producing vis weight sum if vis present """
    if present:
        def impl(shape, array):
            return np.zeros(shape, dtype=array.real.dtype)
    else:
        def impl(shape, array):
            pass

    return njit(nogil=True, cache=True)(impl)


def chan_output_factory(present):
    """ Returns function producing outputs if the array is present """
    if present:
        def impl(shape, array):
            return np.zeros(shape, dtype=array.dtype)
    else:
        def impl(shape, array):
            pass

    return njit(nogil=True, cache=True)(impl)


def vis_add_factory(have_vis, have_weight, have_weight_spectrum):
    """ Returns function adding weighted visibilities to a bin """
    if not have_vis:
        def impl(out_vis, out_weight_sum, in_vis, weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):
            pass
    elif have_weight_spectrum:
        # Always prefer more accurate weight spectrum if we have it
        def impl(out_vis, out_weight_sum, in_vis,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            wt = weight_spectrum[irow, ichan, corr]
            iv = in_vis[irow, ichan, corr] * wt
            out_vis[orow, ochan, corr] += iv
            out_weight_sum[orow, ochan, corr] += wt

    elif have_weight:
        # Otherwise fall back to row weights
        def impl(out_vis, out_weight_sum, in_vis,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            wt = weight[irow, corr]
            iv = in_vis[irow, ichan, corr] * wt
            out_vis[orow, ochan, corr] += iv
            out_weight_sum[orow, ochan, corr] += wt
    else:
        # Natural weights
        def impl(out_vis, out_weight_sum, in_vis,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            iv = in_vis[irow, ichan, corr]
            out_vis[orow, ochan, corr] += iv
            out_weight_sum[orow, ochan, corr] += 1.0

    return njit(nogil=True, cache=True)(impl)


def sigma_spectrum_add_factory(have_vis, have_weight, have_weight_spectrum):
    """ Returns function adding weighted sigma to a bin """
    if not have_vis:
        def impl(out_sigma, out_weight_sum, in_sigma,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):
            pass

    elif have_weight_spectrum:
        # Always prefer more accurate weight spectrum if we have it
        def impl(out_sigma, out_weight_sum, in_sigma,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            # sum(sigma**2 * weight**2)
            wt = weight_spectrum[irow, ichan, corr]
            is_ = in_sigma[irow, ichan, corr]**2 * wt**2
            out_sigma[orow, ochan, corr] += is_
            out_weight_sum[orow, ochan, corr] += wt

    elif have_weight:
        # Otherwise fall back to row weights
        def impl(out_sigma, out_weight_sum, in_sigma,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            # sum(sigma**2 * weight**2)
            wt = weight[irow, corr]
            is_ = in_sigma[irow, ichan, corr]**2 * wt**2
            out_sigma[orow, ochan, corr] += is_
            out_weight_sum[orow, ochan, corr] += wt
    else:
        # Natural weights
        def impl(out_sigma, out_weight_sum, in_sigma,
                 weight, weight_spectrum,
                 orow, ochan, irow, ichan, corr):

            # sum(sigma**2 * weight**2)
            out_sigma[orow, ochan, corr] += in_sigma[irow, ichan, corr]**2
            out_weight_sum[orow, ochan, corr] += 1.0

    return njit(nogil=True, cache=True)(impl)


def chan_add_factory(present):
    """ Returns function for adding data to a bin """
    if present:
        def impl(output, input, orow, ochan, irow, ichan, corr):
            output[orow, ochan, corr] += input[irow, ichan, corr]
    else:
        def impl(output, input, orow, ochan, irow, ichan, corr):
            pass

    return njit(nogil=True, cache=True)(impl)


def vis_normaliser_factory(present):
    if present:
        def impl(vis_out, vis_in, row, chan, corr, weight_sum):
            wsum = weight_sum[row, chan, corr]

            if wsum != 0.0:
                vis_out[row, chan, corr] = vis_in[row, chan, corr] / wsum
    else:
        def impl(vis_out, vis_in, row, chan, corr, weight_sum):
            pass

    return njit(nogil=True, cache=True)(impl)


def sigma_spectrum_normaliser_factory(present):
    if present:
        def impl(sigma_out, sigma_in, row, chan, corr, weight_sum):
            wsum = weight_sum[row, chan, corr]

            if wsum == 0.0:
                return

            # sqrt(sigma**2 * weight**2 / (weight(sum**2)))
            res = np.sqrt(sigma_in[row, chan, corr] / (wsum**2))
            sigma_out[row, chan, corr] = res
    else:
        def impl(sigma_out, sigma_in, row, chan, corr, weight_sum):
            pass

    return njit(nogil=True, cache=True)(impl)


def weight_spectrum_normaliser_factory(present):
    if present:
        def impl(wt_spec_out, wt_spec_in, row, chan, corr):
            wt_spec_out[row, chan, corr] = wt_spec_in[row, chan, corr]
    else:
        def impl(wt_spec_out, wt_spec_in, row, chan, corr):
            pass

    return njit(nogil=True, cache=True)(impl)


def chan_normaliser_factory(present):
    """ Returns function normalising channel data in a bin """
    if present:
        def impl(data_out, data_in, row, chan, corr, bin_size):
            data_out[row, chan, corr] = data_in[row, chan, corr] / bin_size
    else:
        def impl(data_out, data_in, row, chan, corr, bin_size):
            pass

    return njit(nogil=True, cache=True)(impl)


@generated_jit(nopython=True, nogil=True, cache=True)
def shape_or_invalid_shape(array, ndim):
    """ Return array shape tuple or (-1,)*ndim if the array is None """

    try:
        ndim_lit = getattr(ndim, "literal_value")
    except AttributeError:
        raise ValueError("ndim must be a integer literal")

    if is_numba_type_none(array):
        tup = (-1,)*ndim_lit

        def impl(array, ndim):
            return tup
    else:
        def impl(array, ndim):
            return array.shape

    return impl


@njit(nogil=True, cache=True)
def find_chan_corr(chan, corr, shape, chan_idx, corr_idx):
    """
    1. Get channel and correlation from shape if not set and the shape is valid
    2. Check they agree if they already agree

    Parameters
    ----------
    chan : int
        Existing channel size
    corr : int
        Existing correlation size
    shape : tuple
        Array shape tuple
    chan_idx : int
        Index of channel dimension in ``shape``.
    corr_idx : int
        Index of correlation dimension in ``shape``.

    Returns
    -------
    int
        Modified channel size
    int
        Modified correlation size
    """
    if chan_idx != -1:
        array_chan = shape[chan_idx]

        # Corresponds to a None array, ignore
        if array_chan == -1:
            pass
        # chan is not yet set, assign
        elif chan == 0:
            chan = array_chan
        # Check consistency
        elif chan != array_chan:
            raise ValueError("Inconsistent Channel Dimension "
                             "in Input Arrays")

    if corr_idx != -1:
        array_corr = shape[corr_idx]

        # Corresponds to a None array, ignore
        if array_corr == -1:
            pass
        # corr is not yet set, assign
        elif corr == 0:
            corr = array_corr
        # Check consistency
        elif corr != array_corr:
            raise ValueError("Inconsistent Correlation Dimension "
                             "in Input Arrays")

    return chan, corr


@njit(nogil=True, cache=True)
def chan_corrs(vis, flag,
               weight_spectrum, sigma_spectrum,
               chan_freq, chan_width,
               effective_bw, resolution):
    """
    Infer channel and correlation size from input dimensions

    Returns
    -------
    int
        channel size
    int
        correlation size
    """
    vis_shape = shape_or_invalid_shape(vis, 3)
    flag_shape = shape_or_invalid_shape(flag, 3)
    weight_spectrum_shape = shape_or_invalid_shape(weight_spectrum, 3)
    sigma_spectrum_shape = shape_or_invalid_shape(sigma_spectrum, 3)
    chan_freq_shape = shape_or_invalid_shape(chan_freq, 1)
    chan_width_shape = shape_or_invalid_shape(chan_width, 1)
    effective_bw_shape = shape_or_invalid_shape(effective_bw, 1)
    resolution_shape = shape_or_invalid_shape(resolution, 1)

    chan = 0
    corr = 0

    chan, corr = find_chan_corr(chan, corr, vis_shape, 1, 2)
    chan, corr = find_chan_corr(chan, corr, flag_shape, 1, 2)
    chan, corr = find_chan_corr(chan, corr, weight_spectrum_shape, 1, 2)
    chan, corr = find_chan_corr(chan, corr, sigma_spectrum_shape, 1, 2)
    chan, corr = find_chan_corr(chan, corr, chan_freq_shape, 0, -1)
    chan, corr = find_chan_corr(chan, corr, chan_width_shape, 0, -1)
    chan, corr = find_chan_corr(chan, corr, effective_bw_shape, 0, -1)
    chan, corr = find_chan_corr(chan, corr, resolution_shape, 0, -1)

    return chan, corr


def is_chan_flagged_factory(present):
    if present:
        def impl(flag, r, f, c):
            return flag[r, f, c]
    else:
        def impl(flag, r, f, c):
            return False

    return njit(nogil=True, cache=True)(impl)


def set_flagged_factory(present):
    if present:
        def impl(flag, r, f, c):
            flag[r, f, c] = 1
    else:
        def impl(flag, r, f, c):
            pass

    return njit(nogil=True, cache=True)(impl)


_rowchan_output_fields = ["vis", "flag", "weight_spectrum", "sigma_spectrum"]
RowChanAverageOutput = namedtuple("RowChanAverageOutput",
                                  _rowchan_output_fields)


class RowChannelAverageException(Exception):
    pass


@generated_jit(nopython=True, nogil=True, cache=True)
def row_chan_average(row_meta, chan_meta, flag_row=None, weight=None,
                     vis=None, flag=None,
                     weight_spectrum=None, sigma_spectrum=None):

    have_flag_row = not is_numba_type_none(flag_row)
    have_vis = not is_numba_type_none(vis)
    have_flag = not is_numba_type_none(flag)
    have_weight = not is_numba_type_none(weight)
    have_weight_spectrum = not is_numba_type_none(weight_spectrum)
    have_sigma_spectrum = not is_numba_type_none(sigma_spectrum)

    flags_match = matching_flag_factory(have_flag_row)
    is_chan_flagged = is_chan_flagged_factory(have_flag)

    vis_factory = chan_output_factory(have_vis)
    weight_sum_factory = weight_sum_output_factory(have_vis)
    flag_factory = chan_output_factory(have_flag)
    weight_factory = chan_output_factory(have_weight_spectrum)
    sigma_factory = chan_output_factory(have_sigma_spectrum)

    vis_adder = vis_add_factory(have_vis,
                                have_weight,
                                have_weight_spectrum)
    weight_adder = chan_add_factory(have_weight_spectrum)
    sigma_adder = sigma_spectrum_add_factory(have_sigma_spectrum,
                                             have_weight,
                                             have_weight_spectrum)

    vis_normaliser = vis_normaliser_factory(have_vis)
    sigma_normaliser = sigma_spectrum_normaliser_factory(have_sigma_spectrum)
    weight_normaliser = weight_spectrum_normaliser_factory(
                            have_weight_spectrum)

    set_flagged = set_flagged_factory(have_flag)

    dummy_chan_freq = None
    dummy_chan_width = None

    def impl(row_meta, chan_meta, flag_row=None, weight=None,
             vis=None, flag=None,
             weight_spectrum=None, sigma_spectrum=None):

        out_rows = row_meta.time.shape[0]
        nchan, ncorrs = chan_corrs(vis, flag,
                                   weight_spectrum, sigma_spectrum,
                                   dummy_chan_freq, dummy_chan_width,
                                   dummy_chan_width, dummy_chan_width)

        chan_map, out_chans = chan_meta

        out_shape = (out_rows, out_chans, ncorrs)

        vis_avg = vis_factory(out_shape, vis)
        vis_weight_sum = weight_sum_factory(out_shape, vis)
        weight_spectrum_avg = weight_factory(out_shape, weight_spectrum)
        sigma_spectrum_avg = sigma_factory(out_shape, sigma_spectrum)
        sigma_spectrum_weight_sum = sigma_factory(out_shape, sigma_spectrum)

        flagged_vis_avg = vis_factory(out_shape, vis)
        flagged_vis_weight_sum = weight_sum_factory(out_shape, vis)
        flagged_weight_spectrum_avg = weight_factory(out_shape,
                                                     weight_spectrum)
        flagged_sigma_spectrum_avg = sigma_factory(out_shape,
                                                   sigma_spectrum)
        flagged_sigma_spectrum_weight_sum = sigma_factory(out_shape,
                                                          sigma_spectrum)

        flag_avg = flag_factory(out_shape, flag)

        counts = np.zeros(out_shape, dtype=np.uint32)
        flag_counts = np.zeros(out_shape, dtype=np.uint32)

        # Iterate over input rows, accumulating into output rows
        for in_row, out_row in enumerate(row_meta.map):
            # TIME_CENTROID/EXPOSURE case applies here,
            # must have flagged input and output OR unflagged input and output
            if not flags_match(flag_row, in_row, row_meta.flag_row, out_row):
                continue

            for in_chan, out_chan in enumerate(chan_map):
                for corr in range(ncorrs):
                    if is_chan_flagged(flag, in_row, in_chan, corr):
                        # Increment flagged averages and counts
                        flag_counts[out_row, out_chan, corr] += 1

                        vis_adder(flagged_vis_avg, flagged_vis_weight_sum, vis,
                                  weight, weight_spectrum,
                                  out_row, out_chan, in_row, in_chan, corr)
                        weight_adder(flagged_weight_spectrum_avg,
                                     weight_spectrum,
                                     out_row, out_chan, in_row, in_chan, corr)
                        sigma_adder(flagged_sigma_spectrum_avg,
                                    flagged_sigma_spectrum_weight_sum,
                                    sigma_spectrum,
                                    weight,
                                    weight_spectrum,
                                    out_row, out_chan, in_row, in_chan, corr)
                    else:
                        # Increment unflagged averages and counts
                        counts[out_row, out_chan, corr] += 1

                        vis_adder(vis_avg, vis_weight_sum, vis,
                                  weight, weight_spectrum,
                                  out_row, out_chan, in_row, in_chan, corr)
                        weight_adder(weight_spectrum_avg, weight_spectrum,
                                     out_row, out_chan, in_row, in_chan, corr)
                        sigma_adder(sigma_spectrum_avg,
                                    sigma_spectrum_weight_sum,
                                    sigma_spectrum,
                                    weight,
                                    weight_spectrum,
                                    out_row, out_chan, in_row, in_chan, corr)

        for r in range(out_rows):
            for f in range(out_chans):
                for c in range(ncorrs):
                    if counts[r, f, c] > 0:
                        # We have some unflagged samples and
                        # only these are used as averaged output
                        vis_normaliser(vis_avg, vis_avg,
                                       r, f, c,
                                       vis_weight_sum)
                        sigma_normaliser(sigma_spectrum_avg,
                                         sigma_spectrum_avg,
                                         r, f, c,
                                         sigma_spectrum_weight_sum)
                    elif flag_counts[r, f, c] > 0:
                        # We only have flagged samples and
                        # these are used as averaged output
                        vis_normaliser(vis_avg, flagged_vis_avg,
                                       r, f, c,
                                       flagged_vis_weight_sum)
                        sigma_normaliser(sigma_spectrum_avg,
                                         flagged_sigma_spectrum_avg,
                                         r, f, c,
                                         flagged_sigma_spectrum_weight_sum)
                        weight_normaliser(weight_spectrum_avg,
                                          flagged_weight_spectrum_avg,
                                          r, f, c)

                        # Flag the output bin
                        set_flagged(flag_avg, r, f, c)
                    else:
                        raise RowChannelAverageException("Zero-filled bin")

        return RowChanAverageOutput(vis_avg, flag_avg,
                                    weight_spectrum_avg,
                                    sigma_spectrum_avg)

    return impl


_chan_output_fields = ["chan_freq", "chan_width", "effective_bw", "resolution"]
ChannelAverageOutput = namedtuple("ChannelAverageOutput", _chan_output_fields)


@generated_jit(nopython=True, nogil=True, cache=True)
def chan_average(chan_meta, chan_freq=None, chan_width=None,
                 effective_bw=None, resolution=None):
    have_chan_freq = not is_numba_type_none(chan_freq)
    have_chan_width = not is_numba_type_none(chan_width)
    have_effective_bw = not is_numba_type_none(effective_bw)
    have_resolution = not is_numba_type_none(resolution)

    chan_freq_output = chan_output_factory(have_chan_freq)
    chan_width_output = chan_output_factory(have_chan_width)
    effective_bw_output = chan_output_factory(have_effective_bw)
    resolution_output = chan_output_factory(have_resolution)

    chan_freq_normaliser = normaliser_factory(have_chan_freq)

    chan_freq_adder = add_factory(have_chan_freq)
    chan_width_adder = add_factory(have_chan_width)
    effective_bw_adder = add_factory(have_effective_bw)
    resolution_adder = add_factory(have_resolution)

    def impl(chan_meta, chan_freq=None, chan_width=None,
             effective_bw=None, resolution=None):
        chan_map, out_chans = chan_meta

        chan_freq_avg = chan_freq_output(out_chans, chan_freq)
        chan_width_avg = chan_width_output(out_chans, chan_width)
        effective_bw_avg = effective_bw_output(out_chans, effective_bw)
        resolution_avg = resolution_output(out_chans, resolution)
        counts = np.zeros(out_chans, dtype=np.uint32)

        for in_chan, out_chan in enumerate(chan_map):
            counts[out_chan] += 1
            chan_freq_adder(chan_freq_avg, out_chan, chan_freq, in_chan)
            chan_width_adder(chan_width_avg, out_chan, chan_width, in_chan)
            effective_bw_adder(effective_bw_avg, out_chan,
                               effective_bw, in_chan)
            resolution_adder(resolution_avg, out_chan,
                             resolution, in_chan)

        for out_chan in range(out_chans):
            chan_freq_normaliser(chan_freq_avg, out_chan, counts[out_chan])

        return ChannelAverageOutput(chan_freq_avg, chan_width_avg,
                                    effective_bw_avg, resolution_avg)

    return impl


AverageOutput = namedtuple("AverageOutput",
                           ["time", "interval", "flag_row"] +
                           _row_output_fields +
                           _chan_output_fields +
                           _rowchan_output_fields)


@generated_jit(nopython=True, nogil=True, cache=True)
def merge_flags(flag_row, flag):
    have_flag_row = not is_numba_type_none(flag_row)
    have_flag = not is_numba_type_none(flag)

    if have_flag_row and have_flag:
        def impl(flag_row, flag):
            """ Check flag_row and flag agree """
            for r in range(flag.shape[0]):
                all_flagged = True

                for f in range(flag.shape[1]):
                    for c in range(flag.shape[2]):
                        if flag[r, f, c] == 0:
                            all_flagged = False
                            break

                    if not all_flagged:
                        break

                if (flag_row[r] != 0) != all_flagged:
                    raise ValueError("flag_row and flag arrays mismatch")

            return flag_row

    elif have_flag_row and not have_flag:
        def impl(flag_row, flag):
            """ Return flag_row """
            return flag_row

    elif not have_flag_row and have_flag:
        def impl(flag_row, flag):
            """ Construct flag_row from flag """
            new_flag_row = np.empty(flag.shape[0], dtype=flag.dtype)

            for r in range(flag.shape[0]):
                all_flagged = True

                for f in range(flag.shape[1]):
                    for c in range(flag.shape[2]):
                        if flag[r, f, c] == 0:
                            all_flagged = False
                            break

                    if not all_flagged:
                        break

                new_flag_row[r] = (1 if all_flagged else 0)

            return new_flag_row

    else:
        def impl(flag_row, flag):
            return None

    return impl


@generated_jit(nopython=True, nogil=True, cache=True)
def time_and_channel(time, interval, antenna1, antenna2,
                     time_centroid=None, exposure=None, flag_row=None,
                     uvw=None, weight=None, sigma=None,
                     chan_freq=None, chan_width=None,
                     effective_bw=None, resolution=None,
                     vis=None, flag=None,
                     weight_spectrum=None, sigma_spectrum=None,
                     time_bin_secs=1.0, chan_bin_size=1):

    valid_types = (types.misc.Omitted, types.scalars.Float,
                   types.scalars.Integer)

    if not isinstance(time_bin_secs, valid_types):
        raise TypeError("time_bin_secs must be a scalar float")

    valid_types = (types.misc.Omitted, types.scalars.Integer)

    if not isinstance(chan_bin_size, valid_types):
        raise TypeError("chan_bin_size must be a scalar integer")

    def impl(time, interval, antenna1, antenna2,
             time_centroid=None, exposure=None, flag_row=None,
             uvw=None, weight=None, sigma=None,
             chan_freq=None, chan_width=None,
             effective_bw=None, resolution=None,
             vis=None, flag=None,
             weight_spectrum=None, sigma_spectrum=None,
             time_bin_secs=1.0, chan_bin_size=1):

        nchan, ncorrs = chan_corrs(vis, flag,
                                   weight_spectrum, sigma_spectrum,
                                   chan_freq, chan_width,
                                   effective_bw, resolution)

        # Merge flag_row and flag arrays
        flag_row = merge_flags(flag_row, flag)

        # Generate row mapping metadata
        row_meta = row_mapper(time, interval, antenna1, antenna2,
                              flag_row=flag_row, time_bin_secs=time_bin_secs)

        # Generate channel mapping metadata
        chan_meta = channel_mapper(nchan, chan_bin_size)

        # Average row data
        row_data = row_average(row_meta, antenna1, antenna2, flag_row=flag_row,
                               time_centroid=time_centroid, exposure=exposure,
                               uvw=uvw, weight=weight, sigma=sigma)

        # Average channel data
        chan_data = chan_average(chan_meta, chan_freq=chan_freq,
                                 chan_width=chan_width,
                                 effective_bw=effective_bw,
                                 resolution=resolution)

        # Average row and channel data
        row_chan_data = row_chan_average(row_meta, chan_meta,
                                         flag_row=flag_row, weight=weight,
                                         vis=vis, flag=flag,
                                         weight_spectrum=weight_spectrum,
                                         sigma_spectrum=sigma_spectrum)

        # Have to explicitly write it out because numba tuples
        # are highly constrained types
        return AverageOutput(row_meta.time,
                             row_meta.interval,
                             row_meta.flag_row,
                             row_data.antenna1,
                             row_data.antenna2,
                             row_data.time_centroid,
                             row_data.exposure,
                             row_data.uvw,
                             row_data.weight,
                             row_data.sigma,
                             chan_data.chan_freq,
                             chan_data.chan_width,
                             chan_data.effective_bw,
                             chan_data.resolution,
                             row_chan_data.vis,
                             row_chan_data.flag,
                             row_chan_data.weight_spectrum,
                             row_chan_data.sigma_spectrum)

    return impl


AVERAGING_DOCS = DocstringTemplate("""
Averages in time and channel.

Parameters
----------
time : $(array_type)
    Time values of shape :code:`(row,)`.
interval : $(array_type)
    Interval values of shape :code:`(row,)`.
antenna1 : $(array_type)
    First antenna indices of shape :code:`(row,)`
antenna2 : $(array_type)
    Second antenna indices of shape :code:`(row,)`
time_centroid : $(array_type), optional
    Time centroid values of shape :code:`(row,)`
exposure : $(array_type), optional
    Exposure values of shape :code:`(row,)`
flag_row : $(array_type), optional
    Flagged rows of shape :code:`(row,)`.
uvw : $(array_type), optional
    UVW coordinates of shape :code:`(row, 3)`.
weight : $(array_type), optional
    Weight values of shape :code:`(row, corr)`.
sigma : $(array_type), optional
    Sigma values of shape :code:`(row, corr)`.
chan_freq : $(array_type), optional
    Channel frequencies of shape :code:`(chan,)`.
chan_width : $(array_type), optional
    Channel widths of shape :code:`(chan,)`.
effective_bw : $(array_type), optional
    Effective channel bandwidth of shape :code:`(chan,)`.
resolution : $(array_type), optional
    Effective channel resolution of shape :code:`(chan,)`.
vis : $(array_type), optional
    Visibility data of shape :code:`(row, chan, corr)`.
flag : $(array_type), optional
    Flag data of shape :code:`(row, chan, corr)`.
weight_spectrum : $(array_type), optional
    Weight spectrum of shape :code:`(row, chan, corr)`.
sigma_spectrum : $(array_type), optional
    Sigma spectrum of shape :code:`(row, chan, corr)`.
time_bin_secs : float, optional
    Maximum summed interval in seconds to include within a bin.
    Defaults to 1.0.
chan_bin_size : int, optional
    Number of bins to average together.
    Defaults to 1.

Notes
-----

The implementation currently requires unique lexicographical
combinations of (TIME, ANTENNA1, ANTENNA2). This can usually
be achieved by suitably partitioning input data on indexing rows,
DATA_DESC_ID and SCAN_NUMBER in particular.

Returns
-------
namedtuple
    A namedtuple whose entries correspond to the input arrays.
    Output arrays will be ``None`` if the inputs were ``None``.
""")


try:
    time_and_channel.__doc__ = AVERAGING_DOCS.substitute(
                                    array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import logging
from os.path import join as pjoin

import numpy as np

from africanus.constants import minus_two_pi_over_c
from africanus.util.jinja2 import jinja_env
from africanus.rime.phase import PHASE_DELAY_DOCS
from africanus.util.code import memoize_on_key, format_code
from africanus.util.cuda import cuda_function, grids
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
    from cupy.core._scalar import get_typename as _get_typename
    from cupy.cuda.compiler import CompileException
except ImportError:
    pass

log = logging.getLogger(__name__)


def _key_fn(lm, uvw, frequency):
    return (lm.dtype, uvw.dtype, frequency.dtype)


_TEMPLATE_PATH = pjoin("rime", "cuda", "phase.cu.j2")


@memoize_on_key(_key_fn)
def _generate_kernel(lm, uvw, frequency):
    # Floating point output type
    out_dtype = np.result_type(lm, uvw, frequency)

    # Block sizes
    blockdimx = 32 if frequency.dtype == np.float32 else 16
    blockdimy = 32 if uvw.dtype == np.float32 else 16
    block = (blockdimx, blockdimy, 1)

    # Create template
    render = jinja_env.get_template(_TEMPLATE_PATH).render
    name = "phase_delay"

    code = render(kernel_name=name,
                  lm_type=_get_typename(lm.dtype),
                  uvw_type=_get_typename(uvw.dtype),
                  freq_type=_get_typename(frequency.dtype),
                  out_type=_get_typename(out_dtype),
                  sqrt_fn=cuda_function('sqrt', lm.dtype),
                  sincos_fn=cuda_function('sincos', out_dtype),
                  minus_two_pi_over_c=minus_two_pi_over_c,
                  blockdimx=blockdimx,
                  blockdimy=blockdimy).encode('utf-8')

    # Complex output type
    out_dtype = np.result_type(out_dtype, np.complex64)
    return cp.RawKernel(code, name), block, out_dtype


@requires_optional("cupy")
def phase_delay(lm, uvw, frequency):
    kernel, block, out_dtype = _generate_kernel(lm, uvw, frequency)
    grid = grids((frequency.shape[0], uvw.shape[0], 1), block)
    out = cp.empty(shape=(lm.shape[0], uvw.shape[0], frequency.shape[0]),
                   dtype=out_dtype)

    try:
        kernel(grid, block, (lm, uvw, frequency, out))
    except CompileException:
        log.exception(format_code(kernel.code))
        raise

    return out


try:
    phase_delay.__doc__ = PHASE_DELAY_DOCS.substitute(
                                array_type=':class:`cupy.ndarray`')
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import logging
import os
from os.path import join as pjoin
import shutil
import tempfile
from urllib.request import urlopen
from zipfile import ZipFile

try:
    from dask.utils import SerializableLock as Lock
except ImportError:
    from threading import Lock

from africanus.util.appdirs import downloads_dir, include_dir
from africanus.util.files import sha_hash_file

_trove_dir = pjoin(include_dir, "trove")
_trove_url = 'https://github.com/bryancatanzaro/trove/archive/master.zip'
_trove_sha_hash = 'b4a0ac97f23fcc94064a75c5cab9464e4c65f136'
_trove_version_str = 'Current release: v1.8.0 (02/16/2018)'
_trove_version = "master"
_trove_zip_dir = 'trove-' + _trove_version
_trove_download_filename = "trove-" + _trove_version + ".zip"
_trove_header = pjoin(_trove_dir, 'trove', 'trove.cuh')
_trove_readme = pjoin(_trove_dir, 'README.md')
_trove_new_unzipped_path = _trove_dir


log = logging.getLogger(__name__)


class InstallTroveException(Exception):
    pass


def download_trove(archive_file):
    response = urlopen(_trove_url)

    if not os.path.exists(downloads_dir):
        os.makedirs(downloads_dir)

    with open(archive_file, "wb") as f:
        shutil.copyfileobj(response, f)

    return sha_hash_file(archive_file)


def is_trove_installed(readme_filename):
    # Check if the README.md exists
    if (not os.path.exists(readme_filename) or
            not os.path.isfile(readme_filename)):

        reason = "trove readme '{}' does not exist".format(readme_filename)
        return (False, reason)

    return (True, "")


def _install_trove():
    archive = pjoin(downloads_dir, _trove_download_filename)
    should_download = True

    # Check existing archive in the download directory
    # If the hash matches we can avoid downloading
    if os.path.exists(archive) and os.path.isfile(archive):
        if _trove_sha_hash == sha_hash_file(archive):
            should_download = False

    # Download and check the hash
    if should_download:
        sha_hash = download_trove(archive)
        # Compare against our supplied hash
        if _trove_sha_hash != sha_hash:
            msg = ('Hash of file %s downloaded from %s '
                   'is %s and does not match the expected '
                   'hash of %s.') % (
                        _trove_download_filename, _trove_url,
                        _trove_sha_hash, sha_hash)

            raise InstallTroveException(msg)

    # Unzip into include/trove
    with ZipFile(archive, 'r') as zip_file:
        # Remove any existing install
        try:
            shutil.rmtree(_trove_dir, ignore_errors=True)
        except Exception as e:
            raise InstallTroveException("Removing %s failed\n%s" % (
                                      _trove_dir, str(e)))

        try:
            # Unzip into temporary directory
            tmpdir = tempfile.mkdtemp()
            zip_file.extractall(tmpdir)

            unzip_path = pjoin(tmpdir, _trove_zip_dir)

            # Move
            shutil.move(unzip_path, _trove_dir)
        except Exception as e:
            raise InstallTroveException("Extracting %s failed\n%s" % (
                                      archive, str(e)))
        finally:
            shutil.rmtree(tmpdir, ignore_errors=True)

        log.info("NVIDIA trove archive unzipped into '%s'" % _trove_dir)

    # Final check on installation
    there, reason = is_trove_installed(_trove_readme)

    if not there:
        raise InstallTroveException(reason)


_trove_install_lock = Lock()

with _trove_install_lock:
    _trove_installed, _ = is_trove_installed(_trove_readme)


def trove_dir():
    global _trove_installed

    if _trove_installed is False:
        with _trove_install_lock:
            # Double-locking pattern
            if _trove_installed is False:
                _install_trove()
                _trove_installed = True

    return _trove_dir



# -*- coding: utf-8 -*-


import logging
import os
from os.path import join as pjoin
import shutil
import tempfile
from urllib.request import urlopen
from zipfile import ZipFile

try:
    from dask.utils import SerializableLock as Lock
except ImportError:
    from threading import Lock

from africanus.util.appdirs import downloads_dir, include_dir
from africanus.util.files import sha_hash_file

_cub_dir = pjoin(include_dir, "cub")
_cub_url = 'https://github.com/NVlabs/cub/archive/1.8.0.zip'
_cub_sha_hash = '836f523a34c32a7e99fba36b30abfe7a68d41d4b'
_cub_version_str = 'Current release: v1.8.0 (02/16/2018)'
_cub_version = "1.8.0"
_cub_zip_dir = 'cub-' + _cub_version
_cub_download_filename = "cub-" + _cub_version + ".zip"
_cub_header = pjoin(_cub_dir, 'cub', 'cub.cuh')
_cub_readme = pjoin(_cub_dir, 'README.md')
_cub_new_unzipped_path = _cub_dir


log = logging.getLogger(__name__)


class InstallCubException(Exception):
    pass


def download_cub(archive_file):
    response = urlopen(_cub_url)

    if not os.path.exists(downloads_dir):
        os.makedirs(downloads_dir)

    with open(archive_file, "wb") as f:
        shutil.copyfileobj(response, f)

    return sha_hash_file(archive_file)


def is_cub_installed(readme_filename, header_filename, cub_version_str):
    # Check if the cub.h exists
    if (not os.path.exists(header_filename) or
            not os.path.isfile(header_filename)):

        reason = "CUB header '{}' does not exist".format(header_filename)
        return (False, reason)

    # Check if the README.md exists
    if (not os.path.exists(readme_filename) or
            not os.path.isfile(readme_filename)):

        reason = "CUB readme '{}' does not exist".format(readme_filename)
        return (False, reason)

    # Search for the version string, returning True if found
    with open(readme_filename, 'r') as f:
        for line in f:
            if line.find(cub_version_str) != -1:
                return (True, "")

    # Nothing found!
    reason = "CUB version string '{}' not found in '{}'".format(
        cub_version_str, readme_filename)

    return (False, reason)


def _install_cub():
    archive = pjoin(downloads_dir, _cub_download_filename)
    should_download = True

    # Check existing archive in the download directory
    # If the hash matches we can avoid downloading
    if os.path.exists(archive) and os.path.isfile(archive):
        if _cub_sha_hash == sha_hash_file(archive):
            should_download = False

    # Download and check the hash
    if should_download:
        sha_hash = download_cub(archive)
        # Compare against our supplied hash
        if _cub_sha_hash != sha_hash:
            msg = ('Hash of file %s downloaded from %s '
                   'is %s and does not match the expected '
                   'hash of %s.') % (
                        _cub_download_filename, _cub_url,
                        _cub_sha_hash, sha_hash)

            raise InstallCubException(msg)

    # Unzip into include/cub
    with ZipFile(archive, 'r') as zip_file:
        # Remove any existing install
        try:
            shutil.rmtree(_cub_dir, ignore_errors=True)
        except Exception as e:
            raise InstallCubException("Removing %s failed\n%s" % (
                                      _cub_dir, str(e)))

        try:
            # Unzip into temporary directory
            tmpdir = tempfile.mkdtemp()
            zip_file.extractall(tmpdir)

            unzip_path = pjoin(tmpdir, _cub_zip_dir)

            # Move
            shutil.move(unzip_path, _cub_dir)
        except Exception as e:
            raise InstallCubException("Extracting %s failed\n%s" % (
                                      archive, str(e)))
        finally:
            shutil.rmtree(tmpdir, ignore_errors=True)

        log.info("NVIDIA cub archive unzipped into '%s'" % _cub_dir)

    # Final check on installation
    there, reason = is_cub_installed(_cub_readme, _cub_header,
                                     _cub_version_str)

    if not there:
        raise InstallCubException(reason)


_cub_install_lock = Lock()

with _cub_install_lock:
    _cub_installed, _ = is_cub_installed(_cub_readme, _cub_header,
                                         _cub_version_str)


def cub_dir():
    global _cub_installed

    if _cub_installed is False:
        with _cub_install_lock:
            # Double-locking pattern
            if _cub_installed is False:
                _install_cub()
                _cub_installed = True

    return _cub_dir



# -*- coding: utf-8 -*-
from numba import types
import numpy as np

from africanus.util.numba import generated_jit
from africanus.util.docs import DocstringTemplate


def ordinary_spectral_model(I, coeffs, log_poly, freq, ref_freq):
    """ Numpy ordinary polynomial implementation """
    coeffs_idx = np.arange(1, coeffs.shape[1] + 1)
    # (source, chan, coeffs-comp)
    term = (freq[None, :, None] / ref_freq[:, None, None]) - 1.0
    term = term**coeffs_idx[None, None, :]
    term = coeffs[:, None, :]*term
    return I[:, None] + term.sum(axis=2)


def log_spectral_model(I, coeffs, log_poly, freq, ref_freq):
    """ Numpy logarithmic polynomial implementation """
    # No negative flux
    I = np.where(log_poly == False, 1.0, I)  # noqa
    coeffs_idx = np.arange(1, coeffs.shape[1] + 1)
    # (source, chan, coeffs-comp)
    term = np.log(freq[None, :, None] / ref_freq[:, None, None])
    term = term**coeffs_idx[None, None, :]
    term = coeffs[:, None, :]*term
    return np.exp(np.log(I)[:, None] + term.sum(axis=2))


@generated_jit(nopython=True, nogil=True, cache=True)
def _check_log_poly_shape(coeffs, log_poly):
    if isinstance(log_poly, types.npytypes.Array):
        def impl(coeffs, log_poly):
            if coeffs.shape[0] != log_poly.shape[0]:
                raise ValueError("coeffs.shape[0] != log_poly.shape[0]")
    elif isinstance(log_poly, types.scalars.Boolean):
        def impl(coeffs, log_poly):
            pass
    else:
        raise ValueError("log_poly must be ndarray or bool")

    return impl


@generated_jit(nopython=True, nogil=True, cache=True)
def _log_polynomial(log_poly, s):
    if isinstance(log_poly, types.npytypes.Array):
        def impl(log_poly, s):
            return log_poly[s]
    elif isinstance(log_poly, types.scalars.Boolean):
        def impl(log_poly, s):
            return log_poly
    else:
        raise ValueError("log_poly must be ndarray or bool")

    return impl


@generated_jit(nopython=True, nogil=True, cache=True)
def spectra(I, coeffs, log_poly, ref_freq, frequency):
    arg_dtypes = tuple(np.dtype(a.dtype.name) for a
                       in (I, coeffs, ref_freq, frequency))
    dtype = np.result_type(*arg_dtypes)

    def impl(I, coeffs, log_poly, ref_freq, frequency):
        if not (I.shape[0] == coeffs.shape[0] == ref_freq.shape[0]):
            raise ValueError("first dimensions of I, coeffs "
                             "and ref_freq don't match.")

        _check_log_poly_shape(coeffs, log_poly)

        nsrc = I.shape[0]
        nchan = frequency.shape[0]
        ncoeffs = coeffs.shape[1]

        spectral_model = np.empty((nsrc, nchan), dtype=dtype)

        for s in range(nsrc):
            rf = ref_freq[s]

            if _log_polynomial(log_poly, s):
                for f in range(frequency.shape[0]):
                    nu = frequency[f]

                    # Initialise with base polynomial value
                    spectral_model[s, f] = np.log(I[s])

                    for c in range(ncoeffs):
                        term = coeffs[s, c]
                        term *= np.log(nu/rf)**(c + 1)
                        spectral_model[s, f] += term

                    spectral_model[s, f] = np.exp(spectral_model[s, f])
            else:
                for f in range(frequency.shape[0]):
                    nu = frequency[f]

                    # Initialise with base polynomial value
                    spectral_model[s, f] = I[s]

                    for c in range(ncoeffs):
                        term = coeffs[s, c]
                        term *= ((nu/rf) - 1.0)**(c + 1)
                        spectral_model[s, f] += term

        return spectral_model

    return impl


SPECTRA_DOCS = DocstringTemplate(r"""
Produces a spectral model from a polynomial expansion of
a wsclean file model. Depending on how `log_poly` is set
ordinary or logarithmic polynomials are used to produce
the expansion:

.. math::

    & flux(\lambda) =
      I_{0} + \sum\limits_{c=0} \textrm{coeffs}(c)
              ({\lambda/\lambda_{ref}} - 1)^{c+1}
              \\
    & flux(\lambda) =
      \exp \left( \log I_{0} +
              \sum\limits_{c=0} \textrm{coeffs}(c)
              \log({\lambda/\lambda_{ref}})^{c+1}
            \right) \\


See the `WSClean Component List
<https://sourceforge.net/p/wsclean/wiki/ComponentList/>`_
for further details.

Parameters
----------
I : $(array_type)
    flux density in Janskys at the reference frequency
    of shape :code:`(source,)`
coeffs : $(array_type)
    Polynomial coefficients for each source of
    shape :code:`(source, comp)`
log_poly : $(array_type) or bool
    boolean array of shape :code:`(source, )`
    indicating whether logarithmic (True) or ordinary (False)
    polynomials should be used.
ref_freq : $(array_type)
    Source reference frequencies of shape :code:`(source,)`
frequency : $(array_type)
    frequencies of shape :code:`(chan,)`

See Also
--------
africanus.model.wsclean.load

Returns
-------
spectral_model : $(array_type)
    Spectral Model of shape :code:`(source, chan)`
""")

try:
    spectra.__doc__ = SPECTRA_DOCS.substitute(
                            array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import pytest

from africanus.averaging.support import unique_time, unique_baselines
from africanus.averaging.time_and_channel_mapping import (row_mapper,
                                                          channel_mapper)


@pytest.fixture
def time():
    return np.asarray([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])  # noqa


@pytest.fixture
def interval():
    data = np.asarray([1.9, 2.0, 2.1, 1.85, 1.95, 2.0, 2.05, 2.1, 2.05, 1.9])
    return data*0.1


@pytest.fixture
def ant1():
    return np.asarray([0,   0,   1,   0,   0,   1,   2,   0,   0,   1],  # noqa
                      dtype=np.int32)


@pytest.fixture
def ant2():
    return np.asarray([1,   2,   2,   0,   1,   2,   3,   0,   1,   2],  # noqa
                      dtype=np.int32)


def flag_row_factory(nrows, flagged_rows):
    flag_row = np.zeros(nrows, dtype=np.uint8)

    if flagged_rows is not None:
        flag_row[flagged_rows] = 1

    return flag_row


@pytest.mark.parametrize("time_bin_secs", [0.1, 0.2, 1, 2, 4])
@pytest.mark.parametrize("flagged_rows", [None, [0, 1], [2, 4], range(10)])
def test_row_mapper(time, interval, ant1, ant2,
                    flagged_rows, time_bin_secs):
    utime, _, time_inv, _ = unique_time(time)
    ubl, _, bl_inv, _ = unique_baselines(ant1, ant2)
    mask = np.full((ubl.shape[0], utime.shape[0]), -1, dtype=np.int32)

    mask[bl_inv, time_inv] = np.arange(time.size)

    flag_row = flag_row_factory(time.size, flagged_rows)

    ret = row_mapper(time, interval, ant1, ant2,
                     flag_row=flag_row,
                     time_bin_secs=time_bin_secs)

    # For TIME AND INTERVAL, flagged inputs can
    # contribute to unflagged outputs
    new_time = np.zeros_like(ret.time)
    new_interval = np.zeros_like(ret.interval)
    counts = np.zeros(ret.time.shape, dtype=np.uint32)
    np.add.at(new_time, ret.map, time)
    np.add.at(new_interval, ret.map, interval)
    np.add.at(counts, ret.map, 1)

    assert_array_equal(ret.time, new_time / counts)
    assert_array_equal(ret.interval, new_interval)

    # For TIME_CENTROID and EXPOSURE,
    # unflagged inputs only contribute to unflagged outputs and
    # flagged inputs only contribute to flagged outputs

    # Now recalculate time_avg using the row_map
    new_tc = np.zeros_like(ret.time)
    new_exp = np.zeros_like(ret.interval)
    counts = np.zeros(ret.time.shape, dtype=np.uint32)

    sel = flag_row == ret.flag_row[ret.map]
    np.add.at(new_tc, ret.map[sel], time[sel])
    np.add.at(new_exp, ret.map[sel], interval[sel])
    np.add.at(counts, ret.map[sel], 1)

    ant1_avg = np.empty(ret.time.shape, dtype=ant1.dtype)
    ant2_avg = np.empty(ret.time.shape, dtype=ant2.dtype)
    ant1_avg[ret.map[sel]] = ant1[sel]
    ant2_avg[ret.map[sel]] = ant2[sel]

    # Do it a different way
    new_tc2 = np.zeros_like(ret.time)
    new_exp2 = np.zeros_like(ret.interval)
    counts2 = np.zeros(ret.time.shape, dtype=np.uint32)

    for ri, ro in enumerate(ret.map):
        if flag_row[ri] == 1 and ret.flag_row[ro] == 1:
            new_tc2[ro] += time[ri]
            new_exp2[ro] += interval[ri]
            counts2[ro] += 1
        elif flag_row[ri] == 0 and ret.flag_row[ro] == 0:
            new_tc2[ro] += time[ri]
            new_exp2[ro] += interval[ri]
            counts2[ro] += 1

    assert_array_almost_equal(new_tc / counts, new_tc2 / counts2)
    assert_array_almost_equal(new_exp, new_exp2)


def test_channel_mapper():
    chan_map, out_chans = channel_mapper(64, 17)

    uchan, counts = np.unique(chan_map, return_counts=True)

    assert_array_equal(chan_map[0:17], 0)
    assert_array_equal(chan_map[17:34], 1)
    assert_array_equal(chan_map[34:51], 2)
    assert_array_equal(chan_map[51:64], 3)

    assert_array_equal(uchan, [0, 1, 2, 3])
    assert_array_equal(counts, [17, 17, 17, 13])

    assert out_chans == 4



# -*- coding: utf-8 -*-


import warnings

from .parangles_astropy import (have_astropy_parangles,
                                astropy_parallactic_angles)
from .parangles_casa import (have_casa_parangles,
                             casa_parallactic_angles)

_discovered_backends = []

if have_astropy_parangles:
    _discovered_backends.append('astropy')

if have_casa_parangles:
    _discovered_backends.append('casa')


_standard_backends = set(['casa', 'astropy', 'test'])


def parallactic_angles(times, antenna_positions, field_centre,
                       backend='casa'):
    """
    Computes parallactic angles per timestep for the given
    reference antenna position and field centre.

    Parameters
    ----------
    times : :class:`numpy.ndarray`
        Array of Mean Julian Date times in
        *seconds* with shape :code:`(time,)`,
    antenna_positions : :class:`numpy.ndarray`
        Antenna positions of shape :code:`(ant, 3)`
        in *metres* in the *ITRF* frame.
    field_centre : :class:`numpy.ndarray`
        Field centre of shape :code:`(2,)` in *radians*
    backend : {'casa', 'test'}, optional
        Backend to use for calculating the parallactic angles.

        * ``casa`` defers to an implementation
          depending on ``python-casacore``.
          This backend should be used by default.
        * ``test`` creates parallactic angles
          by multiplying the ``times`` and ``antenna_position``
          arrays. It exist solely for testing.

    Returns
    -------
    parallactic_angles : :class:`numpy.ndarray`
        Parallactic angles of shape :code:`(time,ant)`
    """
    if backend not in _standard_backends:
        raise ValueError("'%s' is not one of the "
                         "standard backends '%s'"
                         % (backend, _standard_backends))

    if not field_centre.shape == (2,):
        raise ValueError("Invalid field_centre shape %s" %
                         (field_centre.shape,))

    if backend == 'astropy':
        warnings.warn('astropy backend currently returns the incorrect values')
        return astropy_parallactic_angles(times,
                                          antenna_positions,
                                          field_centre)
    elif backend == 'casa':
        return casa_parallactic_angles(times,
                                       antenna_positions,
                                       field_centre)
    elif backend == 'test':
        return times[:, None]*(antenna_positions.sum(axis=1)[None, :])
    else:
        raise ValueError("Invalid backend %s" % backend)



# -*- coding: utf-8 -*-


import numpy as np

from africanus.util.docs import DocstringTemplate
from africanus.util.numba import is_numba_type_none, generated_jit, njit


JONES_NOT_PRESENT = 0
JONES_1_OR_2 = 1
JONES_2X2 = 2


def _get_jones_types(name, numba_ndarray_type, corr_1_dims, corr_2_dims):
    """
    Determine which of the following three cases are valid:

    1. The array is not present (None) and therefore no Jones Matrices
    2. single (1,) or (2,) dual correlation
    3. (2, 2) full correlation

    Parameters
    ----------
    name: str
        Array name
    numba_ndarray_type: numba.type
        Array numba type
    corr_1_dims: int
        Number of `numba_ndarray_type` dimensions,
        including correlations (first option)
    corr_2_dims: int
        Number of `numba_ndarray_type` dimensions,
        including correlations (second option)

    Returns
    -------
    int
        Enumeration describing the Jones Matrix Type

        - 0 -- Not Present
        - 1 -- (1,) or (2,)
        - 2 -- (2, 2)
    """

    if is_numba_type_none(numba_ndarray_type):
        return JONES_NOT_PRESENT
    if numba_ndarray_type.ndim == corr_1_dims:
        return JONES_1_OR_2
    elif numba_ndarray_type.ndim == corr_2_dims:
        return JONES_2X2
    else:
        raise ValueError("%s.ndim not in (%d, %d)" %
                         (name, corr_1_dims, corr_2_dims))


def jones_mul_factory(have_ddes, have_coh, jones_type, accumulate):
    """
    Outputs a function that multiplies some combination of
    (dde1_jones, baseline_jones, dde2_jones) together.

    Parameters
    ----------
    have_ddes : boolean
        If True, indicates that antenna jones terms are present
    have_coh : boolean
        If True, indicates that baseline jones terms are present
    jones_type : int
        Type of Jones matrix
    accumulate : boolean
        If True, the result of the multiplication is accumulated
        into the output, otherwise, it is assigned

    Notes
    -----
    ``accumulate`` is treated by LLVM as a compile-time constant,
    according to https://numba.pydata.org/numba-doc/latest/glossary.html.

    Therefore in principle, the conditional checks
    involving ``accumulate`` inside the functions should
    be elided by the compiler.


    Returns
    -------
    callable
        jitted numba function performing the Jones Multiply
    """
    ex = ValueError("Invalid Jones Type %s" % jones_type)

    if have_coh and have_ddes:
        if jones_type == JONES_1_OR_2:
            def jones_mul(a1j, blj, a2j, out):
                for c in range(out.shape[0]):
                    if accumulate:
                        out[c] += a1j[c] * blj[c] * np.conj(a2j[c])
                    else:
                        out[c] = a1j[c] * blj[c] * np.conj(a2j[c])

        elif jones_type == JONES_2X2:
            def jones_mul(a1j, blj, a2j, out):
                a2_xx_H = np.conj(a2j[0, 0])
                a2_xy_H = np.conj(a2j[0, 1])
                a2_yx_H = np.conj(a2j[1, 0])
                a2_yy_H = np.conj(a2j[1, 1])

                xx = blj[0, 0] * a2_xx_H + blj[0, 1] * a2_xy_H
                xy = blj[0, 0] * a2_yx_H + blj[0, 1] * a2_yy_H
                yx = blj[1, 0] * a2_xx_H + blj[1, 1] * a2_xy_H
                yy = blj[1, 0] * a2_yx_H + blj[1, 1] * a2_yy_H

                if accumulate:
                    out[0, 0] += a1j[0, 0] * xx + a1j[0, 1] * yx
                    out[0, 1] += a1j[0, 0] * xy + a1j[0, 1] * yy
                    out[1, 0] += a1j[1, 0] * xx + a1j[1, 1] * yx
                    out[1, 1] += a1j[1, 0] * xy + a1j[1, 1] * yy
                else:
                    out[0, 0] = a1j[0, 0] * xx + a1j[0, 1] * yx
                    out[0, 1] = a1j[0, 0] * xy + a1j[0, 1] * yy
                    out[1, 0] = a1j[1, 0] * xx + a1j[1, 1] * yx
                    out[1, 1] = a1j[1, 0] * xy + a1j[1, 1] * yy

        else:
            raise ex
    elif have_ddes and not have_coh:
        if jones_type == JONES_1_OR_2:
            def jones_mul(a1j, a2j, out):
                for c in range(out.shape[0]):
                    if accumulate:
                        out[c] += a1j[c] * np.conj(a2j[c])
                    else:
                        out[c] = a1j[c] * np.conj(a2j[c])

        elif jones_type == JONES_2X2:
            def jones_mul(a1j, a2j, out):
                a2_xx_H = np.conj(a2j[0, 0])
                a2_xy_H = np.conj(a2j[0, 1])
                a2_yx_H = np.conj(a2j[1, 0])
                a2_yy_H = np.conj(a2j[1, 1])

                if accumulate:
                    out[0, 0] += a1j[0, 0] * a2_xx_H + a1j[0, 1] * a2_xy_H
                    out[0, 1] += a1j[0, 0] * a2_yx_H + a1j[0, 1] * a2_yy_H
                    out[1, 0] += a1j[1, 0] * a2_xx_H + a1j[1, 1] * a2_xy_H
                    out[1, 1] += a1j[1, 0] * a2_yx_H + a1j[1, 1] * a2_yy_H
                else:
                    out[0, 0] += a1j[0, 0] * a2_xx_H + a1j[0, 1] * a2_xy_H
                    out[0, 1] += a1j[0, 0] * a2_yx_H + a1j[0, 1] * a2_yy_H
                    out[1, 0] += a1j[1, 0] * a2_xx_H + a1j[1, 1] * a2_xy_H
                    out[1, 1] += a1j[1, 0] * a2_yx_H + a1j[1, 1] * a2_yy_H
        else:
            raise ex
    elif not have_ddes and have_coh:
        if jones_type == JONES_1_OR_2:
            def jones_mul(blj, out):
                for c in range(out.shape[0]):
                    if accumulate:
                        out[c] += blj[c]
                    elif id(blj) == id(out):
                        pass
                    else:
                        out[c] = blj[c]

        elif jones_type == JONES_2X2:
            def jones_mul(blj, out):
                if accumulate:
                    out[0, 0] += blj[0, 0]
                    out[0, 1] += blj[0, 1]
                    out[1, 0] += blj[1, 0]
                    out[1, 1] += blj[1, 1]
                elif id(blj) == id(out):
                    pass
                else:
                    out[0, 0] = blj[0, 0]
                    out[0, 1] = blj[0, 1]
                    out[1, 0] = blj[1, 0]
                    out[1, 1] = blj[1, 1]
        else:
            raise ex
    else:
        # noop
        def jones_mul():
            pass

    return njit(nogil=True)(jones_mul)


def sum_coherencies_factory(have_ddes, have_coh, jones_type):
    """ Factory function generating a function that sums coherencies """
    jones_mul = jones_mul_factory(have_ddes, have_coh, jones_type, True)

    if have_ddes and have_coh:
        def sum_coh_fn(time, ant1, ant2, a1j, blj, a2j, tmin, out):
            for s in range(a1j.shape[0]):
                for r, (ti, a1, a2) in enumerate(zip(time, ant1, ant2)):
                    ti -= tmin
                    for f in range(a1j.shape[3]):
                        jones_mul(a1j[s, ti, a1, f],
                                  blj[s, r, f],
                                  a2j[s, ti, a2, f],
                                  out[r, f])

    elif have_ddes and not have_coh:
        def sum_coh_fn(time, ant1, ant2, a1j, blj, a2j, tmin, out):
            for s in range(a1j.shape[0]):
                for r, (ti, a1, a2) in enumerate(zip(time, ant1, ant2)):
                    ti -= tmin
                    for f in range(a1j.shape[3]):
                        jones_mul(a1j[s, ti, a1, f],
                                  a2j[s, ti, a2, f],
                                  out[r, f])

    elif not have_ddes and have_coh:
        if jones_type == JONES_2X2:
            def sum_coh_fn(time, ant1, ant2, a1j, blj, a2j, tmin, out):
                for s in range(blj.shape[0]):
                    for r in range(blj.shape[1]):
                        for f in range(blj.shape[2]):
                            for c1 in range(blj.shape[3]):
                                for c2 in range(blj.shape[4]):
                                    out[r, f, c1, c2] += blj[s, r, f, c1, c2]
        else:
            def sum_coh_fn(time, ant1, ant2, a1j, blj, a2j, tmin, out):
                for s in range(blj.shape[0]):
                    for r in range(blj.shape[1]):
                        for f in range(blj.shape[2]):
                            for c in range(blj.shape[3]):
                                out[r, f, c] += blj[s, r, f, c]
    else:
        # noop
        def sum_coh_fn(time, ant1, ant2, a1j, blj, a2j, tmin, out):
            pass

    return njit(nogil=True)(sum_coh_fn)


def output_factory(have_ddes, have_coh, have_dies, have_base_vis, out_dtype):
    """ Factory function generating a function that creates function output """
    if have_ddes:
        def output(time_index, dde1_jones, source_coh, dde2_jones,
                   die1_jones, base_vis, die2_jones):
            row = time_index.shape[0]
            chan = dde1_jones.shape[3]
            corrs = dde1_jones.shape[4:]
            return np.zeros((row, chan) + corrs, dtype=out_dtype)
    elif have_coh:
        def output(time_index, dde1_jones, source_coh, dde2_jones,
                   die1_jones, base_vis, die2_jones):
            row = time_index.shape[0]
            chan = source_coh.shape[2]
            corrs = source_coh.shape[3:]
            return np.zeros((row, chan) + corrs, dtype=out_dtype)
    elif have_dies:
        def output(time_index, dde1_jones, source_coh, dde2_jones,
                   die1_jones, base_vis, die2_jones):
            row = time_index.shape[0]
            chan = die1_jones.shape[2]
            corrs = die1_jones.shape[3:]
            return np.zeros((row, chan) + corrs, dtype=out_dtype)
    elif have_base_vis:
        def output(time_index, dde1_jones, source_coh, dde2_jones,
                   die1_jones, base_vis, die2_jones):
            row = time_index.shape[0]
            chan = base_vis.shape[1]
            corrs = base_vis.shape[2:]
            return np.zeros((row, chan) + corrs, dtype=out_dtype)

    else:
        raise ValueError("Insufficient inputs were supplied "
                         "for determining the output shape")

    return njit(nogil=True)(output)


def add_coh_factory(have_bvis):
    if have_bvis:
        def add_coh(base_vis, out):
            out += base_vis
    else:
        # noop
        def add_coh(base_vis, out):
            pass

    return njit(nogil=True)(add_coh)


def apply_dies_factory(have_dies, have_bvis, jones_type):
    """
    Factory function returning a function that applies
    Direction Independent Effects
    """

    # We always "have visibilities", (the output array)
    jones_mul = jones_mul_factory(have_dies, True, jones_type, False)

    if have_dies and have_bvis:
        def apply_dies(time, ant1, ant2,
                       die1_jones, die2_jones,
                       tmin, out):
            # Iterate over rows
            for r, (ti, a1, a2) in enumerate(zip(time, ant1, ant2)):
                ti -= tmin

                # Iterate over channels
                for c in range(out.shape[1]):
                    jones_mul(die1_jones[ti, a1, c], out[r, c],
                              die2_jones[ti, a2, c], out[r, c])

    elif have_dies and not have_bvis:
        def apply_dies(time, ant1, ant2,
                       die1_jones, die2_jones,
                       tmin, out):
            # Iterate over rows
            for r, (ti, a1, a2) in enumerate(zip(time, ant1, ant2)):
                ti -= tmin

                # Iterate over channels
                for c in range(out.shape[1]):
                    jones_mul(die1_jones[ti, a1, c], out[r, c],
                              die2_jones[ti, a2, c],
                              out[r, c])
    else:
        # noop
        def apply_dies(time, ant1, ant2,
                       die1_jones, die2_jones,
                       tmin, out):
            pass

    return njit(nogil=True)(apply_dies)


def _default_none_check(arg):
    return arg is not None


def predict_checks(time_index, antenna1, antenna2,
                   dde1_jones, source_coh, dde2_jones,
                   die1_jones, base_vis, die2_jones,
                   none_check=_default_none_check):

    have_ddes1 = none_check(dde1_jones)
    have_coh = none_check(source_coh)
    have_ddes2 = none_check(dde2_jones)
    have_dies1 = none_check(die1_jones)
    have_bvis = none_check(base_vis)
    have_dies2 = none_check(die2_jones)

    assert time_index.ndim == 1
    assert antenna1.ndim == 1
    assert antenna2.ndim == 1

    if have_ddes1 ^ have_ddes2:
        raise ValueError("Both dde1_jones and dde2_jones "
                         "must be present or absent")

    if have_dies1 ^ have_dies2:
        raise ValueError("Both die1_jones and die2_jones "
                         "must be present or absent")

    have_ddes = have_ddes1 and have_ddes2
    have_dies = have_dies1 and have_dies2

    if have_ddes1 and dde1_jones.ndim not in (5, 6):
        raise ValueError("dde1_jones.ndim %d not in (5, 6)" % dde1_jones.ndim)

    if have_ddes2 and dde2_jones.ndim not in (5, 6):
        raise ValueError("dde2_jones.ndim %d not in (5, 6)" % dde2_jones.ndim)

    if have_ddes and dde1_jones.ndim != dde2_jones.ndim:
        raise ValueError("dde1_jones.ndim != dde2_jones.ndim")

    if have_coh and source_coh.ndim not in (4, 5):
        raise ValueError("source_coh.ndim %d not in (4, 5)" % source_coh.ndim)

    if have_dies1 and die1_jones.ndim not in (4, 5):
        raise ValueError("die1_jones.ndim %d not in (4, 5)" % die1_jones.ndim)

    if have_bvis and base_vis.ndim not in (3, 4):
        raise ValueError("base_vis.ndim %d not in (3, 4)" % base_vis.ndim)

    if have_dies2 and die2_jones.ndim not in (4, 5):
        raise ValueError("die2_jones.ndim %d not in (4, 5)" % die2_jones.ndim)

    if have_dies1 and have_dies2 and die1_jones.ndim != die2_jones.ndim:
        raise ValueError("die1_jones.ndim != die2_jones.ndim")

    expected_sizes = []

    if have_ddes:
        ndim = dde1_jones.ndim
        expected_sizes.append([ndim, ndim - 1, ndim - 2, ndim - 1]),

    if have_coh:
        ndim = source_coh.ndim
        expected_sizes.append([ndim + 1, ndim, ndim - 1, ndim])

    if have_dies:
        ndim = die1_jones.ndim
        expected_sizes.append([ndim + 1, ndim, ndim - 1, ndim])

    if have_bvis:
        ndim = base_vis.ndim
        expected_sizes.append([ndim + 2, ndim + 1, ndim, ndim + 1])

    if not all(expected_sizes[0] == s for s in expected_sizes[1:]):
        raise ValueError("One of the following pre-conditions is broken "
                         "(missing values are ignored):\n"
                         "dde_jones{1,2}.ndim == source_coh.ndim + 1\n"
                         "dde_jones{1,2}.ndim == base_vis.ndim + 2\n"
                         "dde_jones{1,2}.ndim == die_jones{1,2}.ndim + 1")

    return (have_ddes1, have_coh, have_ddes2,
            have_dies1, have_bvis, have_dies2)


@generated_jit(nopython=True, nogil=True, cache=True)
def predict_vis(time_index, antenna1, antenna2,
                dde1_jones=None, source_coh=None, dde2_jones=None,
                die1_jones=None, base_vis=None, die2_jones=None):

    tup = predict_checks(time_index, antenna1, antenna2,
                         dde1_jones, source_coh, dde2_jones,
                         die1_jones, base_vis, die2_jones,
                         lambda x: not is_numba_type_none(x))

    (have_ddes1, have_coh, have_ddes2, have_dies1, have_bvis, have_dies2) = tup

    # Infer the output dtype
    dtype_arrays = (dde1_jones, source_coh, dde2_jones,
                    die1_jones, base_vis, die2_jones)

    out_dtype = np.result_type(*(np.dtype(a.dtype.name)
                                 for a in dtype_arrays
                                 if not is_numba_type_none(a)))

    jones_types = [
        _get_jones_types("dde1_jones", dde1_jones, 5, 6),
        _get_jones_types("source_coh", source_coh, 4, 5),
        _get_jones_types("dde2_jones", dde2_jones, 5, 6),
        _get_jones_types("die1_jones", die1_jones, 4, 5),
        _get_jones_types("base_vis", base_vis, 3, 4),
        _get_jones_types("die2_jones", die2_jones, 4, 5)]

    ptypes = [t for t in jones_types if t != JONES_NOT_PRESENT]

    if not all(ptypes[0] == p for p in ptypes[1:]):
        raise ValueError("Jones Matrix Correlations were mismatched")

    try:
        jones_type = ptypes[0]
    except IndexError:
        raise ValueError("No Jones Matrices were supplied")

    have_ddes = have_ddes1 and have_ddes2
    have_dies = have_dies1 and have_dies2

    # Create functions that we will use inside our predict function
    out_fn = output_factory(have_ddes, have_coh,
                            have_dies, have_bvis, out_dtype)
    sum_coh_fn = sum_coherencies_factory(have_ddes, have_coh, jones_type)
    apply_dies_fn = apply_dies_factory(have_dies, have_bvis, jones_type)
    add_coh_fn = add_coh_factory(have_bvis)

    def _predict_vis_fn(time_index, antenna1, antenna2,
                        dde1_jones=None, source_coh=None, dde2_jones=None,
                        die1_jones=None, base_vis=None, die2_jones=None):

        # Get the output shape
        out = out_fn(time_index, dde1_jones, source_coh, dde2_jones,
                     die1_jones, base_vis, die2_jones)

        # Minimum time index, used to normalise within function
        tmin = time_index.min()

        # Sum coherencies if any
        sum_coh_fn(time_index, antenna1, antenna2,
                   dde1_jones, source_coh, dde2_jones,
                   tmin, out)

        # Add base visibilities to the output, if any
        add_coh_fn(base_vis, out)

        # Apply direction independent effects, if any
        apply_dies_fn(time_index, antenna1, antenna2,
                      die1_jones, die2_jones,
                      tmin, out)

        return out

    return _predict_vis_fn


@generated_jit(nopython=True, nogil=True, cache=True)
def apply_gains(time_index, antenna1, antenna2,
                die1_jones, corrupted_vis, die2_jones):

    def impl(time_index, antenna1, antenna2,
             die1_jones, corrupted_vis, die2_jones):
        return predict_vis(time_index, antenna1, antenna2,
                           die1_jones=die1_jones,
                           base_vis=corrupted_vis,
                           die2_jones=die2_jones)

    return impl


PREDICT_DOCS = DocstringTemplate(r"""
Multiply Jones terms together to form model visibilities according
to the following formula:

.. math::


    V_{pq} = G_{p} \left(
        B_{pq} + \sum_{s} E_{ps} X_{pqs} E_{qs}^H
        \right) G_{q}^H

where for antenna :math:`p` and :math:`q`, and source :math:`s`:


- :math:`B_{{pq}}` represent base coherencies.
- :math:`E_{{ps}}` represents Direction-Dependent Jones terms.
- :math:`X_{{pqs}}` represents a coherency matrix (per-source).
- :math:`G_{{p}}` represents Direction-Independent Jones terms.

Generally, :math:`E_{ps}`, :math:`G_{p}`, :math:`X_{pqs}`
should be formed by using the `RIME API <rime-api-anchor_>`_ functions
and combining them together with :func:`~numpy.einsum`.

**Please read the Notes**

Notes
-----
* Direction-Dependent terms (dde{1,2}_jones) and
  Independent (die{1,2}_jones) are optional,
  but if one is present, the other must be present.
* The inputs to this function involve ``row``, ``time``
  and ``ant`` (antenna) dimensions.
* Each ``row`` is associated with a pair of antenna Jones matrices
  at a particular timestep via the
  ``time_index``, ``antenna1`` and ``antenna2`` inputs.
* The ``row`` dimension must be an increasing partial order in time.
$(extra_notes)


Parameters
----------
time_index : $(array_type)
    Time index used to look up the antenna Jones index
    for a particular baseline with shape :code:`(row,)`.
    Obtainable via $(get_time_index).
antenna1 : $(array_type)
    Antenna 1 index used to look up the antenna Jones
    for a particular baseline.
    with shape :code:`(row,)`.
antenna2 : $(array_type)
    Antenna 2 index used to look up the antenna Jones
    for a particular baseline.
    with shape :code:`(row,)`.
dde1_jones : $(array_type), optional
    :math:`E_{ps}` Direction-Dependent Jones terms for the first antenna.
    shape :code:`(source,time,ant,chan,corr_1,corr_2)`
source_coh : $(array_type), optional
    :math:`X_{pqs}` Direction-Dependent Coherency matrix for the baseline.
    with shape :code:`(source,row,chan,corr_1,corr_2)`
dde2_jones : $(array_type), optional
    :math:`E_{qs}` Direction-Dependent Jones terms for the second antenna.
    This is usually the same array as ``dde1_jones`` as this
    preserves the symmetry of the RIME. ``predict_vis`` will
    perform the conjugate transpose internally.
    shape :code:`(source,time,ant,chan,corr_1,corr_2)`
die1_jones : $(array_type), optional
    :math:`G_{ps}` Direction-Independent Jones terms for the
    first antenna of the baseline.
    with shape :code:`(time,ant,chan,corr_1,corr_2)`
base_vis : $(array_type), optional
    :math:`B_{pq}` base coherencies, added to source coherency summation
    *before* multiplication with `die1_jones` and `die2_jones`.
    shape :code:`(row,chan,corr_1,corr_2)`.
die2_jones : $(array_type), optional
    :math:`G_{ps}` Direction-Independent Jones terms for the
    second antenna of the baseline.
    This is usually the same array as ``die1_jones`` as this
    preserves the symmetry of the RIME. ``predict_vis`` will
    perform the conjugate transpose internally.
    shape :code:`(time,ant,chan,corr_1,corr_2)`
$(extra_args)

Returns
-------
visibilities : $(array_type)
    Model visibilities of shape :code:`(row,chan,corr_1,corr_2)`
""")


try:
    predict_vis.__doc__ = PREDICT_DOCS.substitute(
                            array_type=":class:`numpy.ndarray`",
                            get_time_index=":code:`np.unique(time, "
                                           "return_inverse=True)[1]`",
                            extra_args="",
                            extra_notes="")
except AttributeError:
    pass


APPLY_GAINS_DOCS = DocstringTemplate(r"""
Apply gains to corrupted visibilities in order to recover
the true visibilities.

Thin wrapper around $(wrapper_func).

.. math::


    V_{pq} = G_{p} C_{pq} G_{q}^H


Parameters
----------
time_index : $(array_type)
    Time index used to look up the antenna Jones index
    for a particular baseline
    with shape :code:`(row,)`.
antenna1 : $(array_type)
    Antenna 1 index used to look up the antenna Jones
    for a particular baseline
    with shape :code:`(row,)`.
antenna2 : $(array_type)
    Antenna 2 index used to look up the antenna Jones
    for a particular baseline
    with shape :code:`(row,)`.
gains1 : $(array_type), optional
    :math:`G_{ps}` Gains for the first antenna of the baseline.
    with shape :code:`(time,ant,chan,corr_1,corr_2)`
corrupted_vis : $(array_type), optiona
    :math:`B_{pq}` corrupted visibilities.
    with shape :code:`(row,chan,corr_1,corr_2)`.
gains2 : $(array_type), optional
    :math:`G_{ps}` Gains for the second antenna of the baseline
    with shape :code:`(time,ant,chan,corr_1,corr_2)`.

Returns
-------
true_vis : $(array_type)
    True visibilities of shape :code:`(row,chan,corr_1,corr_2)`
""")

try:
    apply_gains.__doc__ = APPLY_GAINS_DOCS.substitute(
                        array_type=":class:`numpy.ndarray`",
                        wrapper_func=":func:`~africanus.rime.predict_vis`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


STOKES_TYPES = [
    "Undefined",
    "I",
    "Q",
    "U",
    "V",
    "RR",
    "RL",
    "LR",
    "LL",
    "XX",
    "XY",
    "YX",
    "YY",
    "RX",
    "RY",
    "LX",
    "LY",
    "XR",
    "XL",
    "YR",
    "YL",
    "PP",
    "PQ",
    "QP",
    "QQ",
    "RCircular",
    "LCircular",
    "Linear",
    "Ptotal",
    "Plinear",
    "PFtotal",
    "PFlinear",
    "Pangle"]
"""
List of stokes types as defined in
Measurement Set 2.0 and Stokes.h in casacore:
https://casacore.github.io/casacore/classcasacore_1_1Stokes.html
"""


STOKES_TYPE_MAP = {k: i for i, k in enumerate(STOKES_TYPES)}
"""
Map of stokes type to enumeration as defined in
Measurement Set 2.0 and Stokes.h in casacore:
https://casacore.github.io/casacore/classcasacore_1_1Stokes.html
"""

STOKES_ID_MAP = {v: k for k, v in STOKES_TYPE_MAP.items()}
"""
Map of stokes ID to stokes type string as defined in
Measurement Set 2.0 and Stokes.h in casacore:
https://casacore.github.io/casacore/classcasacore_1_1Stokes.html
"""



# -*- coding: utf-8 -*-


import numpy as np

from africanus.util.requirements import requires_optional

try:
    import pyrap.measures
    import pyrap.quanta as pq
except ImportError as e:
    casa_import_error = e
    have_casa_parangles = False
else:
    casa_import_error = None
    have_casa_parangles = True

    # Create a measures server
    meas_serv = pyrap.measures.measures()


@requires_optional('pyrap.measures', 'pyrap.quanta', casa_import_error)
def casa_parallactic_angles(times, antenna_positions, field_centre,
                            zenith_frame='AZEL'):
    """
    Computes parallactic angles per timestep for the given
    reference antenna position and field centre.
    """

    # Create direction measure for the zenith
    zenith = meas_serv.direction(zenith_frame, '0deg', '90deg')

    # Create position measures for each antenna
    reference_positions = [meas_serv.position(
                                'itrf',
                                *(pq.quantity(x, 'm') for x in pos))
                           for pos in antenna_positions]

    # Compute field centre in radians
    fc_rad = meas_serv.direction('J2000', *(pq.quantity(f, 'rad')
                                            for f in field_centre))

    return np.asarray([
        # Set current time as the reference frame
        meas_serv.do_frame(meas_serv.epoch("UTC", pq.quantity(t, "s")))
        and
        [   # Set antenna position as the reference frame
            meas_serv.do_frame(rp)
            and
            meas_serv.posangle(fc_rad, zenith).get_value("rad")
            for rp in reference_positions
        ]
        for t in times])



# -*- coding: utf-8 -*-


import numpy as np
from africanus.gps.utils import abs_diff


def exponential_squared(x, xp, sigmaf, l, pspec=False):
    """
    Create exponential squared covariance
    function between :math:`D` dimensional
    vectors :math:`x` and :math:`x_p` i.e.

    .. math::
        k(x, x_p) = \\sigma_f^2 \\exp\\left(-\\frac{(x-x_p)^2}{2l^2}\\right)

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        Array of shape :code:`(N, D)`.
    xp : :class:`numpy.ndarray`
        Array of shape :code:`(Np, D)`.
    sigmaf : float
        The signal variance hyper-parameter
    l : float
        The length scale hyper-parameter

    Returns
    -------
    K : :class:`numpy.ndarray`
        Array of shape :code:`(N, Np)`
    """
    if pspec:
        N, D = x.shape
        if D != 1:
            raise(NotImplementedError, "Only 1D pspecs supported")
        if (x != xp).any():
            raise(ValueError, "pspec only defined if x = xp")
        x = x.squeeze()
        delx = x[1] - x[0]
        if (x[1::] - x[0:-1] != delx).any():
            raise(ValueError, "pspec only defined on regular grid")
        s = np.fft.fftshift(np.fft.fftfreq(N, d=delx))
        return np.sqrt(2*np.pi*l)*sigmaf**2.0*np.exp(-l**2*s**2/2.0)
    else:
        xxp = abs_diff(x, xp)
        return sigmaf**2*np.exp(-xxp**2/(2*l**2))



# -*- coding: utf-8 -*-


from africanus.model.spectral.spec_model import (
                                        spectral_model as np_spectral_model,
                                        SPECTRAL_MODEL_DOC)
from africanus.util.requirements import requires_optional

try:
    import dask.array as da
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


def _wrapper(stokes, spi, ref_freq, frequencies, base=None):
    return np_spectral_model(stokes, spi[0], ref_freq, frequencies, base=base)


@requires_optional("dask.array", opt_import_error)
def spectral_model(stokes, spi, ref_freq, frequencies, base=0):
    if len(spi.chunks[1]) != 1:
        raise ValueError("Chunking along the spi dimension unsupported")

    pol_dim = () if stokes.ndim == 1 else ("pol",)

    return da.blockwise(_wrapper, ("source", "chan",) + pol_dim,
                        stokes, ("source",) + pol_dim,
                        spi, ("source", "spi") + pol_dim,
                        ref_freq, ("source",),
                        frequencies, ("chan",),
                        base=base,
                        dtype=stokes.dtype)


try:
    spectral_model.__doc__ = SPECTRAL_MODEL_DOC.substitute(
                                array_type=":class:`dask.array.Array`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_equal
import pytest

from africanus.averaging.support import (unique_baselines, unique_time)


@pytest.fixture
def time():
    return np.asarray([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])  # noqa


@pytest.fixture
def ant1():
    return np.asarray([0,   0,   1,   0,   0,   1,   2,   0,   0,   1],  # noqa
                      dtype=np.int32)


@pytest.fixture
def ant2():
    return np.asarray([1,   2,   2,   0,   1,   2,   3,   0,   1,   2],  # noqa
                      dtype=np.int32)


@pytest.fixture
def vis():
    def _vis(row, chan, fcorrs):
        return (np.arange(row*chan*fcorrs, dtype=np.float32) +
                np.arange(1, row*chan*fcorrs+1, dtype=np.float32)*1j)

    return _vis


def test_unique_time(time):
    # Reverse time to test that sort works
    time = np.flipud(time)

    utime, idx, inv, counts = unique_time(time)
    assert_array_equal(utime, [1.0, 2.0, 3.0])
    assert_array_equal(utime[inv], time)
    assert_array_equal(time[idx], utime)
    assert_array_equal(counts, [3, 4, 3])


def test_unique_baselines(ant1, ant2):
    # Reverse ant1, ant2 to test that sort works
    ant1 = np.flipud(ant1)
    ant2 = np.flipud(ant2)

    test_bl = np.stack([ant1, ant2], axis=1)

    bl, idx, inv, counts = unique_baselines(ant1, ant2)
    assert_array_equal(bl, [[0, 0], [0, 1], [0, 2], [1, 2], [2, 3]])
    assert_array_equal(bl[inv], test_bl)
    assert_array_equal(test_bl[idx], bl)
    assert_array_equal(counts, [2, 3, 1, 3, 1])



# -*- coding: utf-8 -*-


import importlib

from decorator import decorate

from africanus.util.docs import on_rtd
from africanus.util.testing import in_pytest, force_missing_pkg_exception


def _missing_packages(fn, packages, import_errors):
    if len(import_errors) > 0:
        import_err_str = "\n".join((str(e) for e in import_errors))
        return ("%s requires installation of "
                "the following packages: %s.\n"
                "%s" % (fn, packages, import_err_str))
    else:
        return ("%s requires installation of the following packages: %s. "
                % (fn, tuple(packages)))


class MissingPackageException(Exception):
    pass


def requires_optional(*requirements):
    """
    Decorator returning either the original function,
    or a dummy function raising a
    :class:`MissingPackageException` when called,
    depending on whether the supplied ``requirements``
    are present.

    If packages are missing and called within a test, the
    dummy function will call :func:`pytest.skip`.

    Used in the following way:

    .. code-block:: python

        try:
            from scipy import interpolate
        except ImportError as e:
            # https://stackoverflow.com/a/29268974/1611416, pep 3110 and 344
            scipy_import_error = e
        else:
            scipy_import_error = None

        @requires_optional('scipy', scipy_import_error)
        def function(*args, **kwargs):
            return interpolate(...)


    Parameters
    ----------
    requirements : iterable of string, None or ImportError
        Sequence of package names required by the decorated function.
        ImportError exceptions (or None, indicating their absence)
        may also be supplied and will be immediately re-raised within
        the decorator. This is useful for tracking down problems
        in user import logic.

    Returns
    -------
    callable
        Either the original function if all ``requirements``
        are available or a dummy function that throws
        a :class:`MissingPackageException` or skips a pytest.
    """
    # Return a bare wrapper if we're on readthedocs
    if on_rtd():
        def _function_decorator(fn):
            def _wrapper(*args, **kwargs):
                pass

            return decorate(fn, _wrapper)

        return _function_decorator

    have_requirements = True
    missing_requirements = []
    honour_pytest_marker = True
    actual_imports = []
    import_errors = []

    # Try imports
    for requirement in requirements:
        # Ignore
        if requirement is None:
            continue
        # Reraise any supplied ImportErrors
        elif isinstance(requirement, ImportError):
            import_errors.append(requirement)
        # An actual package, try to import it
        elif isinstance(requirement, str):
            try:
                importlib.import_module(requirement)
            except ImportError:
                missing_requirements.append(requirement)
                have_requirements = False
            else:
                actual_imports.append(requirement)
        # We should force exceptions, even if we're in a pytest test case
        elif requirement == force_missing_pkg_exception:
            honour_pytest_marker = False
        # Just wrong
        else:
            raise TypeError("requirements must be "
                            "None, strings or ImportErrors. "
                            "Received %s" % requirement)

    # Requested requirement import succeeded, but there were user
    # import errors that we now re-raise
    if have_requirements and len(import_errors) > 0:
        raise ImportError("Successfully imported %s "
                          "but the following user-supplied "
                          "ImportErrors ocurred: \n%s" %
                          (actual_imports,
                           '\n'.join((str(e) for e in import_errors))))

    def _function_decorator(fn):
        # We have requirements, return the original function
        if have_requirements:
            return fn

        # We don't have requirements, produce a failing wrapper
        def _wrapper(*args, **kwargs):
            """ Empty docstring """

            # We're running test cases
            if honour_pytest_marker and in_pytest():
                try:
                    import pytest
                except ImportError as e:
                    raise ImportError("Marked as in a pytest "
                                      "test case, but pytest cannot "
                                      "be imported! %s" % str(e))
                else:
                    msg = _missing_packages(
                        fn.__name__, missing_requirements, import_errors)
                    pytest.skip(msg)
            # Raise the exception
            else:
                msg = _missing_packages(
                    fn.__name__, missing_requirements, import_errors)
                raise MissingPackageException(msg)

        return decorate(fn, _wrapper)

    return _function_decorator



# -*- coding: utf-8 -*-


try:
    from dask.utils import SerializableLock as Lock
except ImportError:
    from threading import Lock


__run_marker = {'in_pytest': False}
__run_marker_lock = Lock()


# Tag indicating that missing packages should generate an
# exception, regardless of the 'in_pytest' marker
# Used for testing exception raising behaviour
force_missing_pkg_exception = object()


def in_pytest():
    """ Return True if we're marked as executing inside pytest """
    with __run_marker_lock:
        return __run_marker['in_pytest']


def mark_in_pytest(in_pytest=True):
    """ Mark if we're in a pytest run """
    if not type(in_pytest) == bool:
        raise TypeError('in_pytest %s is not a boolean' % in_pytest)

    with __run_marker_lock:
        __run_marker['in_pytest'] = in_pytest



# -*- coding: utf-8 -*-


import numpy as np


def abs_diff(x, xp):
    """
    Gets matrix of differences between
    :math:`D`-dimensional vectors x and xp
    i.e.

    .. math::
        X_{ij} = |x_i - x_j|

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        Array of inputs of shape :code:`(N, D)`.
    xp : :class:`numpy.ndarray`
        Array of inputs of shape :code:`(Np, D)`.

    Returns
    -------
    XX : :class:`numpy.ndarray`
        Array of differences of shape :code:`(N, Np)`.

    """
    try:
        N, D = x.shape
        Np, D = xp.shape
    except Exception:
        N = x.size
        D = 1
        Np = xp.size
        x = np.reshape(x, (N, D))
        xp = np.reshape(xp, (Np, D))
    xD = np.zeros([D, N, Np])
    xpD = np.zeros([D, N, Np])
    for i in range(D):
        xD[i] = np.tile(x[:, i], (Np, 1)).T
        xpD[i] = np.tile(xp[:, i], (N, 1))
    return np.linalg.norm(xD - xpD, axis=0)



# -*- coding: utf-8 -*-


import numpy as np
import pytest

from africanus.util.code import format_code


def test_cuda_shuffle_transpose():
    cp = pytest.importorskip("cupy")
    jinja2 = pytest.importorskip("jinja2")

    _TEMPLATE = jinja2.Template("""
    #include <cupy/carray.cuh>

    #define debug {{debug}}

    extern "C" __global__ void kernel(
        const CArray<{{type}}, 2> input,
        CArray<{{type}}, 2> output)
    {
        const ptrdiff_t & nvis = input.shape()[0];
        int v = blockIdx.x*blockDim.x + threadIdx.x;
        int lane_id = threadIdx.x & ({{warp_size}} - 1);

        if(v >= nvis)
            { return; }

        // Input correlation handled by this thread
        int mask = __activemask();


        {{type}} loads[{{corrs}}];
        {{type}} values[{{corrs}}];

        {% for corr in range(corrs) %}
        loads[{{corr}}] = input[v + {{corr}}*nvis];
        {%- endfor %}

        __syncthreads();

        if(debug)
        {
            if(threadIdx.x == 0)
                { printf("mask %d\\n", mask); }

            printf("[%d] %d %d %d %d\\n",
                   lane_id,
                   loads[0], loads[1],
                   loads[2], loads[3]);

            if(threadIdx.x == 0)
                { printf("\\n"); }
        }


        // Tranpose forward
        #pragma unroll ({{corrs}})
        for(int corr=0; corr < {{corrs}}; ++corr)
        {
            int src_corr = ({{corrs}} - corr + lane_id) % {{corrs}};
            int dest_corr = (lane_id + corr) % {{corrs}};
            int src_lane = (lane_id / {{corrs}})*{{corrs}} + dest_corr;

            values[dest_corr] = __shfl_sync(mask, loads[src_corr],
                                     src_lane, {{warp_size}});
        }

        // Copy
        #pragma unroll ({{corrs}})
        for(int corr=0; corr < {{corrs}}; ++corr)
        {
            loads[corr] = values[corr];
        }

        // Transpose backward
        #pragma unroll ({{corrs}})
        for(int corr=0; corr < {{corrs}}; ++corr)
        {
            int src_corr = ({{corrs}} - corr + lane_id) % {{corrs}};
            int dest_corr = (lane_id + corr) % {{corrs}};
            int src_lane = (lane_id / {{corrs}})*{{corrs}} + dest_corr;

            values[dest_corr] = __shfl_sync(mask, loads[src_corr],
                                     src_lane, {{warp_size}});
        }


        __syncthreads();

        if(debug)
        {
            if(threadIdx.x == 0)
                { printf("\\n"); }

            printf("[%d] %d %d %d %d\\n",
                   lane_id,
                   values[0], values[1],
                   values[2], values[3]);
        }


        {% for corr in range(corrs) %}
        output[v + {{corr}}*nvis] = values[{{corr}}];
        {%- endfor %}
    }
    """)

    nvis = 32
    ncorrs = 4
    dtype = np.int32

    dtypes = {
        np.float32: 'float',
        np.float64: 'double',
        np.int32: 'int',
    }

    code = _TEMPLATE.render(type=dtypes[dtype], warp_size=32,
                            corrs=ncorrs, debug="false").encode("utf-8")
    kernel = cp.RawKernel(code, "kernel")

    inputs = cp.arange(nvis*ncorrs, dtype=dtype).reshape(nvis, ncorrs)
    outputs = cp.empty_like(inputs)
    args = (inputs, outputs)
    block = (256, 1, 1)
    grid = tuple((d + b - 1) // b for d, b in zip((nvis, 1, 1), block))

    try:
        kernel(grid, block, args)
    except cp.cuda.compiler.CompileException:
        print(format_code(kernel.code))
        raise

    np.testing.assert_array_almost_equal(cp.asnumpy(inputs),
                                         cp.asnumpy(outputs))
    return

    # Dead code
    print(grid, block)
    print("\n")
    print(inputs)
    print(outputs)


def register_assign_cycles(N, case=0):
    """
    Determine cycles that stem from performing
    an in-place assignment of the elements of an array.
    In the following. we cannot naively assign the source index
    to the dest index,
    If we assigned the value at index 0 to index 1, the assignment
    from index 1 to index 3 would be invalid, for example:

    src:  [3, 0, 2, 1]
    dest: [0, 1, 2, 3]

    assignment cycles can be broken by using a
    temporary variable to store the contents of a source index
    until dependent assignments have been performed.

    In this way, the minimum number of registers can be used
    to perform the in-place assignment.

    Returns
    -------
    list of lists of tuples
        For example, `[[(0, 2), (2, 0)], [1, 3], [3, 1]]`

    """
    dest = range(N)
    src = [(N - case + n) % N for n in dest]

    deps = {d: s for d, s in zip(dest, src) if d != s}

    for di, d in enumerate(dest):
        si = src.index(d)
        if si > di:
            deps[si] = di

    cycles = []

    while len(deps) > 0:
        k, v = deps.popitem()
        cycle = [(k, v)]

        while True:
            try:
                k = v
                v = deps.pop(k)
            except KeyError:
                # Check that the last key we're trying
                # to get is the first one in the cycle
                assert k == cycle[0][0]
                break

            cycle.append((k, v))

        cycles.append(cycle)

    return cycles


class CupyTemplatingException(Exception):
    def __init__(self, msg):
        super(CupyTemplatingException, self).__init__(msg)


def throw_helper(msg):
    raise CupyTemplatingException(msg)


@pytest.mark.parametrize("ncorrs", [1, 2, 4, 8])
def test_cuda_shuffle_transpose_2(ncorrs):
    cp = pytest.importorskip("cupy")
    jinja2 = pytest.importorskip("jinja2")

    # Implement a warp transpose using Kepler's register shuffle instructions
    # as described in del Mundo's
    # `Towards a performance-portable FFT library for heterogeneous computing`
    # https://doi.org/10.1145/2597917.2597943
    # https://homes.cs.washington.edu/~cdel/papers/cf14-fft.pdf
    # The poster is especially informative
    # https://homes.cs.washington.edu/~cdel/posters/073113-on-efficacy-shuffle-sc2013.pdf
    # and
    # `Enabling Efficient Intra-Warp Communication for
    #  Fourier Transforms in a Many-Core Architecture.`
    # https://homes.cs.washington.edu/~cdel/papers/sc13-shuffle-abstract.pdf
    # http://sc13.supercomputing.org/sites/default/files/PostersArchive/spost142.html

    _TEMPLATE = jinja2.Template("""
    #include <cupy/carray.cuh>

    {%- if (corrs < 1 or (corrs.__and__(corrs - 1) != 0)) %}
    {{ throw("corrs must be 1 or a power of 2") }}
    {%- endif %}

    {% macro warp_transpose(var_name, var_type, var_length, tmp_name="tmp") %}
    {% if var_length > 1 %}
        {
            int mask = __activemask();
            int case_id = threadIdx.x & {{var_length - 1}};
            {{var_type}} {{tmp_name}};  // For variable swaps

            // Horizontal (inter-thread) Rotation
            int addr = case_id;
            {%- for case in range(var_length) %}
            {{var_name}}[{{case}}] = __shfl_sync(mask, {{var_name}}[{{case}}], addr, {{var_length}});
            {%- if not loop.last %}
            addr = __shfl_sync(mask, addr, (case_id + 1) & {{var_length - 1}}, {{var_length}});
            {%- endif %}
            {%- endfor %}

            // Vertical (intra-thread) Rotation
            {%- for case in range(var_length) %}
            // Case {{case}}
            {%- set cycles = register_assign_cycles(corrs, case) %}
            {%- for cycle in cycles %}
            {%- set cstart = cycle[0][0] %}
            {{tmp_name}} = {{var_name}}[{{cstart}}];
            {%- for dest, src in cycle %}
            {%- set src_var = tmp_name if cstart == src else var_name + "[" + src|string + "]" %}
            {{var_name}}[{{dest}}] = case_id == {{case}} ? {{src_var}} : {{var_name}}[{{dest}}];
            {%- endfor %}
            {%- endfor %}
            {%- endfor %}

            // Horizontal (inter-thread) Rotation
            addr = ({{var_length}} - case_id) & {{var_length - 1}};
            {%- for case in range(var_length) %}
            {{var_name}}[{{case}}] = __shfl_sync(mask, {{var_name}}[{{case}}], addr, {{var_length}});
            {%- if not loop.last %}
            addr = __shfl_sync(mask, addr, (case_id + {{var_length - 1}}) & {{var_length - 1}}, {{var_length}});
            {%- endif %}
            {%- endfor %}
        }
    {%- endif %}
    {%- endmacro %}

    {%- set width = corrs %}

    extern "C" __global__ void kernel(
        const CArray<{{type}}, 2> input,
        CArray<{{type}}, 2> output)
    {
        const ptrdiff_t & nvis = input.shape()[0];
        int v = blockIdx.x*blockDim.x + threadIdx.x;

        if(v >= nvis)
            { return; }

        // Array to hold our variables
        {{type}} values[{{corrs}}];

        {% for corr in range(corrs) %}
        values[{{corr}}] = input[v + {{corr}}*nvis];
        {%- endfor %}

        if({{debug}})
        {
            if(threadIdx.x == 0)
                { printf("mask %d\\n", __activemask()); }

            printf("[%d] %d %d %d %d\\n",
                   threadIdx.x & {{warp_size - 1}},
                   values[0], values[1],
                   values[2], values[3]);

            if(threadIdx.x == 0)
                { printf("\\n"); }
        }

        {{ warp_transpose("values", type, corrs) }}
        {{ warp_transpose("values", type, corrs) }}

        if({{debug}})
        {
            if(threadIdx.x == 0)
                { printf("\\n"); }

            printf("[%d] %d %d %d %d\\n",
                   threadIdx.x & {{warp_size - 1}},
                   values[0], values[1],
                   values[2], values[3]);
        }

        {% for corr in range(corrs) %}
        output[v + {{corr}}*nvis] = values[{{corr}}];
        {%- endfor %}
    }
    """)  # noqa

    nvis = 32
    dtype = np.int32

    dtypes = {
        np.float32: 'float',
        np.float64: 'double',
        np.int32: 'int',
    }

    code = _TEMPLATE.render(type=dtypes[dtype],
                            throw=throw_helper,
                            register_assign_cycles=register_assign_cycles,
                            warp_size=32,
                            corrs=ncorrs,
                            debug="false").encode("utf-8")
    kernel = cp.RawKernel(code, "kernel")

    inputs = cp.arange(nvis*ncorrs, dtype=dtype).reshape(nvis, ncorrs)
    outputs = cp.empty_like(inputs)
    args = (inputs, outputs)
    block = (256, 1, 1)
    grid = tuple((d + b - 1) // b for d, b in zip((nvis, 1, 1), block))

    try:
        kernel(grid, block, args)
    except cp.cuda.compiler.CompileException:
        print(format_code(kernel.code))
        raise

    np.testing.assert_array_almost_equal(cp.asnumpy(inputs),
                                         cp.asnumpy(outputs))
    return

    # Dead code
    print(grid, block)
    print("\n")
    print(inputs)
    print(outputs)



# -*- coding: utf-8 -*-


import numpy as np

_array_types = [np.ndarray]

try:
    import dask.array as da
except ImportError:
    pass
else:
    _array_types.append(da.Array)

try:
    import cupy as cp
except ImportError:
    pass
else:
    _array_types.append(cp.ndarray)

_array_types = tuple(_array_types)

cuda_fns = {
    np.dtype(np.float32): {
        'abs': 'fabsf',
        'cos': 'cosf',
        'floor': 'floorf',
        'make2': 'make_float2',
        'max': 'fmaxf',
        'min': 'fminf',
        'rsqrt': 'rsqrtf',
        'sqrt': 'sqrtf',
        'sin': 'sinf',
        'sincos': 'sincosf',
        'sincospi': 'sincospif',
    },
    np.dtype(np.float64): {
        'abs': 'fabs',
        'cos': 'cos',
        'floor': 'floor',
        'make2': 'make_double2',
        'max': 'fmax',
        'min': 'fmin',
        'rsqrt': 'rsqrt',
        'sin': 'sin',
        'sincos': 'sincos',
        'sincospi': 'sincospi',
        'sqrt': 'sqrt',
    },
}


numpy_to_cuda_type_map = {
    np.dtype('int8'): "char",
    np.dtype('uint8'): "unsigned char",
    np.dtype('int16'): "short",
    np.dtype('uint16'): "unsigned short",
    np.dtype('int32'): "int",
    np.dtype('uint32'): "unsigned int",
    np.dtype('float32'): "float",
    np.dtype('float64'): "double",
    np.dtype('complex64'): "float2",
    np.dtype('complex128'): "double2"
}

# Also map the types
numpy_to_cuda_type_map.update({k.type: v
                               for k, v
                               in numpy_to_cuda_type_map.items()})


def grids(dims, blocks):
    """
    Determine the grid size, given space dimensions sizes and blocks

    Parameters
    ----------
    dims : tuple of ints
        `(x, y, z)` tuple

    Returns
    -------
    tuple
        `(x, y, z)` grid size tuple
    """
    if not len(dims) == 3:
        raise ValueError("dims must be an (x, y, z) tuple. "
                         "CUDA dimension ordering is inverted compared "
                         "to NumPy")

    if not len(blocks) == 3:
        raise ValueError("blocks must be an (x, y, z) tuple. "
                         "CUDA dimension ordering is inverted compared "
                         "to NumPy")

    return tuple((d + b - 1) // b for d, b in zip(dims, blocks))


def cuda_function(function_name, dtype):
    try:
        type_map = cuda_fns[dtype]
    except KeyError:
        raise ValueError("No registered functions for type %s" % dtype)

    try:
        return type_map[function_name]
    except KeyError:
        raise ValueError("Unknown CUDA function %s" % function_name)


def cuda_type(dtype):
    if isinstance(dtype, _array_types):
        dtype = dtype.dtype

    try:
        return numpy_to_cuda_type_map[dtype]
    except KeyError:
        raise ValueError("No registered map for type %s" % dtype)



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_almost_equal
import pickle
import pytest

from africanus.gridding.nifty.dask import (grid, degrid, dirty, model,
                                           grid_config)


def rf(*a, **kw):
    return np.random.random(*a, **kw)


def rc(*a, **kw):
    return rf(*a, **kw) + 1j*rf(*a, **kw)


def test_dask_nifty_gridder():
    """ Only tests that we can call it and create a dirty image """
    dask = pytest.importorskip('dask')
    da = pytest.importorskip('dask.array')
    _ = pytest.importorskip('nifty_gridder')

    row = (16,)*8
    chan = (32,)
    corr = (4,)
    nx = 1026
    ny = 1022

    nrow = sum(row)
    nchan = sum(chan)
    ncorr = sum(corr)

    # Random UV data
    uvw = rf(size=(nrow, 3)).astype(np.float64)*128
    vis = rf(size=(nrow, nchan, ncorr)).astype(np.complex128)
    freq = np.linspace(.856e9, 2*.856e9, nchan)
    flag = np.zeros(vis.shape, dtype=np.uint8)
    flag = np.random.randint(0, 2, vis.shape, dtype=np.uint8).astype(np.bool)
    weight = rf(vis.shape).astype(np.float64)

    da_vis = da.from_array(vis, chunks=(row, chan, corr))
    da_uvw = da.from_array(uvw, chunks=(row, 3))
    da_freq = da.from_array(freq, chunks=chan)
    da_flag = da.from_array(flag, chunks=(row, chan, corr))
    da_weight = da.from_array(weight, chunks=(row, chan, corr))

    gc = grid_config(nx, ny, 2e-13, 2.0, 2.0)

    # Standard fan reduction
    g = grid(da_vis, da_uvw, da_flag, da_weight, da_freq, gc)
    d1 = dirty(g, gc)

    grid_shape = (gc.object.Nu(), gc.object.Nv(), ncorr)
    dirty_shape = (gc.object.Nxdirty(), gc.object.Nydirty(), ncorr)

    assert g.shape == grid_shape
    assert d1.shape == dirty_shape == (nx, ny, ncorr)

    # Stream reduction (single stream)
    g = grid(da_vis, da_uvw, da_flag, da_weight, da_freq, gc, streams=1)
    d2 = dirty(g, gc)

    assert g.shape == grid_shape
    assert d2.shape == dirty_shape == (nx, ny, ncorr)

    # Stream reductino (three streams)
    g = grid(da_vis, da_uvw, da_flag, da_weight, da_freq, gc, streams=3)
    d3 = dirty(g, gc)

    assert g.shape == grid_shape
    assert d3.shape == dirty_shape == (nx, ny, ncorr)

    # All three techniques produce similar results
    d1, d2, d3 = dask.compute(d1, d2, d3)
    assert_array_almost_equal(d1, d2)
    assert_array_almost_equal(d2, d3)


def test_dask_nifty_degridder():
    """ Only tests that we can call it and create some visibilities """
    da = pytest.importorskip('dask.array')
    _ = pytest.importorskip('nifty_gridder')

    row = (16, 16, 16, 16)
    chan = (32,)
    corr = (4,)

    nrow = sum(row)
    nchan = sum(chan)
    ncorr = sum(corr)
    nx = 1026
    ny = 1022

    gc = grid_config(nx, ny, 2e-13, 2.0, 2.0)

    # Random UV data
    uvw = rf(size=(nrow, 3)).astype(np.float64)*128
    freq = np.linspace(.856e9, 2*.856e9, nchan)
    flag = np.zeros((nrow, nchan, ncorr), dtype=np.bool)
    weight = np.ones((nrow, nchan, ncorr), dtype=np.float64)
    image = rc(size=(nx, ny, ncorr)).astype(np.complex128)

    da_uvw = da.from_array(uvw, chunks=(row, 3))
    da_freq = da.from_array(freq, chunks=chan)
    da_flag = da.from_array(flag, chunks=(row, chan, corr))
    da_weight = da.from_array(weight, chunks=(row, chan, corr))
    da_image = da.from_array(image, chunks=(nx, ny, 1))

    da_grid_vis = model(da_image, gc)
    da_vis = degrid(da_grid_vis, da_uvw, da_flag, da_weight, da_freq, gc)
    vis = da_vis.compute()
    assert vis.shape == da_vis.shape


def test_pickle_gridder_config():
    gc = grid_config(512, 1024, 5e-13, 1.3, 2.0)
    gc2 = pickle.loads(pickle.dumps(gc))
    assert gc is not gc2
    assert gc.object.Nxdirty() == gc2.object.Nxdirty() == 512
    assert gc.object.Nydirty() == gc2.object.Nydirty() == 1024
    assert gc.object.Epsilon() == gc2.object.Epsilon() == 5e-13
    assert gc.object.Pixsize_x() == gc2.object.Pixsize_x() == 1.3
    assert gc.object.Pixsize_y() == gc2.object.Pixsize_y() == 2.0



# -*- coding: utf-8 -*-


from functools import reduce
from operator import mul

import numpy as np

from africanus.gridding.simple.gridding import (grid as np_grid_fn,
                                                degrid as np_degrid_fn)
from africanus.util.docs import mod_docs
from africanus.util.requirements import requires_optional

try:
    import dask.array as da
except ImportError as e:
    da_import_error = e
else:
    da_import_error = None


# Unfortunately necessary to introduce an extra dim
# for blockwise to work properly
def _grid_fn(vis, uvw, flags, weights, ref_wave, convolution_filter,
             cell_size, nx, ny):
    return np_grid_fn(vis[0], uvw[0], flags[0], weights[0],
                      ref_wave[0], convolution_filter,
                      cell_size,
                      nx=nx, ny=ny)[None, :]


@requires_optional('dask.array', da_import_error)
def grid(vis, uvw, flags, weights, ref_wave,
         convolution_filter, cell_size, nx=1024, ny=1024):
    """ Documentation below """

    # Creation correlation dimension strings for each correlation
    corrs = tuple('corr-%d' % i for i in range(len(vis.shape[2:])))

    # Get grids, stacked by row
    grids = da.core.blockwise(_grid_fn, ("row", "ny", "nx") + corrs,
                              vis, ("row", "chan") + corrs,
                              uvw, ("row", "(u,v,w)"),
                              flags, ("row", "chan") + corrs,
                              weights, ("row", "chan") + corrs,
                              ref_wave, ("chan",),
                              new_axes={"ny": ny, "nx": nx},
                              adjust_chunks={"row": 1},
                              convolution_filter=convolution_filter,
                              cell_size=cell_size, ny=ny, nx=nx,
                              dtype=vis.dtype)

    # Sum grids over the row dimension to produce (ny, nx, corr_1, corr_2)
    return grids.sum(axis=0)


@requires_optional('dask.array', da_import_error)
def degrid(grid, uvw, weights, ref_wave, convolution_filter, cell_size):
    """ Documentation below """

    grid_flat_corrs = reduce(mul, grid.shape[2:])
    weight_flat_corrs = reduce(mul, weights.shape[2:])

    assert grid_flat_corrs == weight_flat_corrs
    assert uvw.shape[0] == weights.shape[0]
    assert weights.shape[1] == ref_wave.shape[0]

    # Creation correlation dimension strings for each correlation
    corrs = tuple('corr-%d' % i for i in range(len(grid.shape[2:])))

    return da.core.blockwise(np_degrid_fn, ("row", "chan") + corrs,
                             grid, ("ny", "nx") + corrs,
                             uvw, ("row", "(u,v,w)"),
                             weights, ("row", "chan") + corrs,
                             ref_wave, ("chan",),
                             concatenate=True,
                             convolution_filter=convolution_filter,
                             cell_size=cell_size,
                             dtype=np.complex64)


grid.__doc__ = mod_docs(np_grid_fn.__doc__,
                        [(":class:`numpy.ndarray`",
                            ":class:`dask.array.Array`"),
                         ("np.ones_like", "da.ones_like"),
                         ("np.zeros_like", "da.zeros_like")])

degrid.__doc__ = mod_docs(np_degrid_fn.__doc__,
                          [(":class:`numpy.ndarray`",
                            ":class:`dask.array.Array`"),
                           ("np.ones_like", "da.ones_like"),
                           ("np.zeros_like", "da.zeros_like")])



# -*- coding: utf-8 -*-


try:
    import dask.array as da
except ImportError as e:
    dask_import_error = e
else:
    dask_import_error = None


from africanus.util.requirements import requires_optional
from africanus.coordinates.coordinates import (radec_to_lmn as np_radec_to_lmn,
                                               radec_to_lm as np_radec_to_lm,
                                               lmn_to_radec as np_lmn_to_radec,
                                               lm_to_radec as np_lm_to_radec,
                                               RADEC_TO_LMN_DOCS,
                                               LMN_TO_RADEC_DOCS)


def _radec_to_lmn(radec, phase_centre):
    return np_radec_to_lmn(radec[0], phase_centre[0] if phase_centre else None)


@requires_optional('dask.array', dask_import_error)
def radec_to_lmn(radec, phase_centre=None):
    phase_centre_dims = ("radec",) if phase_centre is not None else None

    return da.core.blockwise(_radec_to_lmn, ("source", "lmn"),
                             radec, ("source", "radec"),
                             phase_centre, phase_centre_dims,
                             new_axes={"lmn": 3},
                             dtype=radec.dtype)


def _lmn_to_radec(lmn, phase_centre):
    return np_lmn_to_radec(lmn[0], phase_centre)


@requires_optional('dask.array', dask_import_error)
def lmn_to_radec(lmn, phase_centre=None):
    phase_centre_dims = ("radec",) if phase_centre is not None else None

    return da.core.blockwise(_lmn_to_radec, ("source", "radec"),
                             lmn, ("source", "lmn"),
                             phase_centre, phase_centre_dims,
                             new_axes={"radec": 2},
                             dtype=lmn.dtype)


def _radec_to_lm(radec, phase_centre):
    return np_radec_to_lm(radec[0], phase_centre[0] if phase_centre else None)


@requires_optional('dask.array', dask_import_error)
def radec_to_lm(radec, phase_centre=None):
    phase_centre_dims = ("radec",) if phase_centre is not None else None

    return da.core.blockwise(_radec_to_lm, ("source", "lm"),
                             radec, ("source", "radec"),
                             phase_centre, phase_centre_dims,
                             new_axes={"lm": 2},
                             dtype=radec.dtype)


def _lm_to_radec(lm, phase_centre):
    return np_lm_to_radec(lm[0], phase_centre)


@requires_optional('dask.array', dask_import_error)
def lm_to_radec(lm, phase_centre=None):
    phase_centre_dims = ("radec",) if phase_centre is not None else None

    return da.core.blockwise(_lm_to_radec, ("source", "radec"),
                             lm, ("source", "lm"),
                             phase_centre, phase_centre_dims,
                             new_axes={"radec": 2},
                             dtype=lm.dtype)


try:
    radec_to_lmn.__doc__ = RADEC_TO_LMN_DOCS.substitute(
                                lm_components="3",
                                array_type=":class:`dask.array.Array`")
    radec_to_lm.__doc__ = RADEC_TO_LMN_DOCS.substitute(
                                lm_components="2",
                                array_type=":class:`dask.array.Array`")
    lmn_to_radec.__doc__ = LMN_TO_RADEC_DOCS.substitute(
                                lm_name="lmn", lm_components="3",
                                array_type=":class:`dask.array.Array`")
    lm_to_radec.__doc__ = LMN_TO_RADEC_DOCS.substitute(
                                lm_name="lm", lm_components="2",
                                array_type=":class:`dask.array.Array`")


except AttributeError:
    pass



# -*- coding: utf-8 -*-


from africanus.testing.beam_factory import beam_factory

import pytest


@pytest.mark.parametrize("pol_type", ["linear", "circular"])
def test_beam_factory(tmp_path, pol_type):
    fits = pytest.importorskip('astropy.io.fits')
    schema = tmp_path / "test_beam_$(corr)_$(reim).fits"

    filenames = beam_factory(schema=schema,
                             npix=15,
                             polarisation_type=pol_type)

    for corr, (re_file, im_file) in filenames.items():
        with fits.open(re_file), fits.open(im_file):
            pass



# -*- coding: utf-8 -*-


from africanus.model.wsclean.spec_model import (spectra as np_spectra,
                                                SPECTRA_DOCS)
from africanus.util.requirements import requires_optional

try:
    import dask.array as da
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


def _wrapper(stokes, spi, log_si, ref_freq, frequency):
    return np_spectra(stokes, spi[0], log_si, ref_freq, frequency)


@requires_optional('dask.array', opt_import_error)
def spectra(stokes, spi, log_si, ref_freq, frequency):
    corrs = tuple("corr-%d" % i for i in range(len(stokes.shape[1:])))
    log_si_schema = None if isinstance(log_si, bool) else ("source",)

    return da.blockwise(_wrapper, ("source", "chan") + corrs,
                        stokes, ("source",) + corrs,
                        spi, ("source", "spi") + corrs,
                        log_si, log_si_schema,
                        ref_freq, ("source",),
                        frequency, ("chan",),
                        dtype=stokes.dtype)


try:
    spectra.__doc__ = SPECTRA_DOCS.substitute(
                            array_type=":class:`dask.array.Array`")
except AttributeError:
    pass
